"""
File Purpose: misc. tools for reading directly from eppic files
"""
import collections
import datetime
import os
import re
import pandas as pd
import xarray as xr

from ...errors import (
    ValueLoadingError,
    FileContentsError, FileContentsConflictError, FileContentsMissingError,
    InputError,
)
from ...tools import (
    attempt_literal_eval, read_idl_params_file,
    InDir, find_jobfiles,
    RESULT_MISSING,
)

''' --------------------- read eppic.i file --------------------- '''

def read_eppic_i_file(filename="eppic.i", *, eval=True):
    '''Parse EPPIC input from the file, returning (dict of global vars, dict of dists).
    dict of global vars has {varname: value} for all pairs not inside a distribution.
    dict of dists has {N: {varname: value} for all pairs in dist} for each dist;
        for each varname in a dist, ensure varname ends with str(N) (append it if necessary).
        the N's will all be stored here as ints.

    Arguments:
    filename: string, default 'eppic.i'
        file to read from.
    eval: bool, default True
        whether to attempt to evaluate the values,
        using ast.literal_eval (safer but less flexible than eval).
        if True, try to evaluate values but use strings if evaluation fails.
        if False, values will remain as strings.

    Notes:
    - file formatting:
        - semicolons (;) are used for comments. (idl format)
        - ignore blank lines & lines that don't assign a variable (missing '=')
        - ignores all leading & trailing whitespace in vars & values.
    - special var: "dist{N}", where N is any integer (or empty string).
        - after finding this, start storing lines in result["dist{dist}"]
            - e.g. "dist1=1" --> result["dist1"][var] = value.
                same behavior if "dist=1".
            - additionally, N will be appended to var names if it is not already there,
                e.g. "dist1=1", "var=7" --> result["dist1"]["var1"] = 7
        - if "dist{N}" appears AGAIN, use it as the new dist; store lines in result["dist{N}"]
    - special var: "fndist"
        - if found, stop storing lines in result["dist{dist}"]
    '''
    
    globs = dict()
    dists = dict()
    
    var_value_pairs = read_idl_params_file(filename,
                                           eval=False,  # eval below instead
                                           as_tuples=True)  # maintain order via tuples not dict.

    target = globs  # << this is where we are currently storing values.
    for (var, value) in var_value_pairs:
        # handle special var: "dist{N}"  (N might be the empty string)
        match = re.match(r'dist(\d*)', var)
        if match:
            # get "N"
            N, = match.groups()
            if N == '':
                N = value
            elif int(N) != int(value):
                raise NotImplementedError(f'found "distN = M", but N!=M. N={N!r}, M={value!r}')
            intN = int(N)
            strN = str(N)
            # "dist{N}" is found. Start storing lines in dists[int(N)]
            if intN in dists:
                raise NotImplementedError(f'"dist=N" (or distN=N) appears multiple times. N={N}')
            else:
                target = dict()
                dists[intN] = target
            continue
        # else:
        if eval:
            value = attempt_literal_eval(value)
        # handle special var: "fndist"
        if var == 'fndist':
            target = globs
            target[var] = value
            continue
        # handle other vars
        if eval:
            value = attempt_literal_eval(value)
        if target is not globs:  # if in a dist right now, ensure var ends with N
            if not var.endswith(strN):
                var = f'{var}{strN}'
        if var in target and target[var]!=value:  # Make warning if duplicate value.
            print(f'Warning: {repr(var)} appears in multiple lines, with different values!')
        # <<< actually store a value
        target[var] = value

    return (globs, dists)

def infer_eppic_dist_names(filename="eppic.i", *, missing_ok=True):
    '''return dict of {distnumber: name} for all dists in eppic.i file.
    name will be None if failed to infer the name.
    name is inferred from comment lines above the lines like dist=N.

    if not missing_ok, name must be inferred for all dists, otherwise crash.

    Comments with names have a few different possible formats:
    1) "; [anything] distribution: name".
        Possibly using ' ' or '-' instead of ':', and possibly using 'dist' instead of distribution.
        Also, possibly have another comment afterwards, e.g. "; [anything] dist - name  ; [anything]".
    2) "; [anything] --- name --- [anything]",
        with at least 3 '-' on either side, and nonzero whitespace between the '-' and the name.
    '''
    # read the file lines.
    with open(filename,"r") as file:
        lines = file.read().splitlines()
    # remove leading whitespace on all lines
    lines = [line.lstrip() for line in lines]
    # find location of lines like dist = N; capture dist number. (line might look like distN = N, instead)
    dist_lines = dict()  # {dist number: line number}
    for i, line in enumerate(lines):
        match = re.fullmatch(r'\s*dist(\d+)?\s*=\s*(\d+).*', line)
        if match:
            _, distN_as_str = match.groups()
            distN = int(distN_as_str)
            dist_lines[distN] = i
    # find location of earliest comment above dist line but after any other assignment line,
    #    and with less than 2 fully empty lines between it and the dist line.
    name_lines = dict()   # {dist number: line number of line with name}
    for distN, i in dist_lines.items():
        k = None
        nempty = 0
        for isub in range(1, i):
            j = i - isub
            if lines[j].startswith(';'):
                k = j
                nempty = 0
            elif '=' in lines[j]:
                break  # found a line with an equal sign; don't look for any higher-up comment lines.
            elif lines[j] == '':
                nempty += 1
                if nempty >= 2:
                    break  # found 2 empty lines in a row; don't look for any higher-up comment lines
        name_lines[distN] = k   # will be None if no name line found.
        if not missing_ok and k is None:
            raise ValueLoadingError(f'failed to find comment which might have name for dist {distN}')
    # infer names from comments. see docstring for details.
    dist_names = {i: None for i in dist_lines.keys()}  # {dist number: name}.  if name not found, use None.
    for i in name_lines:
        if name_lines[i] is not None:
            line = lines[name_lines[i]]
            # check if match format 1
            match = re.fullmatch(r';.*dist(ribution)?\s*[:-]?\s*([^;]+);?.*', line)
            if match:
                _, name = match.groups()
                dist_names[i] = name.strip()
                continue
            # check if match format 2
            match = re.fullmatch(r';.*---+\s+([^;]+)\s+---+.*', line)
            if match:
                name, = match.groups()
                dist_names[i] = name.strip()
                continue
            # no match
            if not missing_ok:
                errmsg = (f'failed to match name_lines[{i}] to any name-comment-line format;\n'
                          '(Acceptable formats include: "; --- name ---" and also "; [anything] dist[ribution] - name")\n'
                          f'at line {name_lines[i]} in file {os.path.abspath(filename)!r};\n'
                          f'line = {repr(line)}')
                raise ValueLoadingError(errmsg)
    return dist_names

''' --------------------- update eppic.i file --------------------- '''
# i.e., make a new one based on the existing one, possibly with some parameters changed.

def update_eppic_i_file(src="eppic.i", dst="eppic_updated.i", values=dict(), *,
                        exists_ok=False, missing_values_ok=True, comment='previously={old}'):
    '''update eppic.i file from src, writing to dst.
    values: dict
        {varname: value} for all pairs to update.
        Can also provide {dist number: dict of {varname: value} pairs to update for this dist}
            (inside of dist number dicts, varname can end with dist number or not;
            assume it does end with dist number in the file itself though)
    exists_ok: bool, default False
        whether it is okay for dst to already exist.
    missing_values_ok: bool
        whether it is okay for some of the varnames in values to not exist in eppic.i
    comment: None or str
        if not None, append comment as a comment to all updated lines.
            (actually appends: '  ; {comment}'.)
        Only appends if line doesn't already end with comment.
        Also, this comment will be hit by .format(old=old value)

    Note: this routine isn't smart enough to add new varnames to eppic.i;
        it's only smart enough to update the values at existing varnames.

    return abspath to dst.
    '''
    # check that dst doesn't exist already
    if not exists_ok and os.path.exists(dst):
        errmsg = (f'dst file {dst!r} already exists!\n(abspath: {os.path.abspath(dst)!r})\n'
                  'Set exists_ok=True to enable overwriting existing file.')
        raise FileExistsError(errmsg)
    # get string to use for eppic.i file updated with values:
    updated_eppic_i_str = get_updated_eppic_i_str(src, values, missing_values_ok=missing_values_ok, comment=comment)
    # update values; using write mode such that previous file will be overwritten.
    with open(dst, 'w') as file:
        file.write(updated_eppic_i_str)
    # return abspath to dst
    return os.path.abspath(dst)

def get_updated_eppic_i_str(src="eppic.i", values=dict(), *, missing_values_ok=True, comment='previously={old}'):
    '''get str to use for updated eppic.i file from src, updated with these values.
    See help(update_eppic_i_file) for more details.
    '''
    # read lines from eppic.i file
    with open(src, 'r') as file:
        lines = file.read().splitlines()
    # get values to plug directly into eppic.i file (flatten any distribution dicts in values)
    # (note: pop values from values_flat after using them; crash if any remain unpopped but not missing_values_ok.)
    values_flat = _get_update_eppic_i_values_flat(values)
    # update lines, retaining all formatting and comments.
    # uses re to do smart matching, handling the fact that some value-setting lines might have comments;
    # lines look like:
    #    varname = value   ; possibly with a comment
    # Additionally, append comment as a comment (with leading ';') to end of lines with changed values.
    comment = '' if comment is None else f'  ; {comment}'
    for i, line in enumerate(lines):
        match = re.fullmatch(r'(\s*)(\w+)(\s*)(=)(\s*)(.*?)(\s*;.*)?', line)
        if match:
            s0, varname0, s1, _eq_, s2, value0, comment0 = match.groups()
            if varname0 in values_flat:
                new_value = str(values_flat.pop(varname0))
                if comment0 is None:
                    comment0 = ''
                lines[i] = f'{s0}{varname0}{s1}{_eq_}{s2}{new_value}{comment0}'
                # add comment if line doesn't end with it already, AND value was changed.
                value_same = (str(value0) == str(new_value))
                if not value_same:  # strs are unequal; maybe represent same number though.
                    value_same = (attempt_literal_eval(value0) == attempt_literal_eval(new_value))
                if not value_same:
                    comment_add = comment.format(old=value0)
                    if not line.rstrip().endswith(comment_add.rstrip()):
                        lines[i] = f'{lines[i]}{comment_add}'
    # check that all values were used
    if not missing_values_ok and len(values_flat) > 0:
        errmsg = (f'Some values were not used; key(s) not found: {list(values_flat.keys())}'
                    f'\nabspath to src = {os.path.abspath(src)!r}')
        raise ValueError(errmsg)
    return '\n'.join(lines)

def _get_update_eppic_i_values_flat(values=dict()):
    '''return "flat" dict of values to use for updating eppic.i file.
    This is a dict of {varname: value} for all pairs to update, in the original file.
    in result, varname for vars in dists will end with dist number.

    values: dict
        {varname: value} for all pairs to update.
        Can also provide {dist number: dict of {varname: value} pairs to update for this dist}
            (inside of dist number dicts, varname can end with dist number or not;
            assume it does end with dist number in the file itself though)
    '''
    result = dict()
    for varname, value in values.items():
        if isinstance(value, dict):
            N = str(varname)   # dist number.
            assert N.isdigit(), f'Expected dist key to look like a number, but got {varname!r}.'
            for varname2, value2 in value.items():
                # append dist number to varname2, if it doesn't end with it already:
                if not varname2.endswith(N):
                    varname2 = f'{varname2}{N}'
                result[varname2] = value2
        else:
            result[varname] = value
    return result


''' --------------------- read eppic snaps info --------------------- '''

def read_eppic_snaps_info(dir=os.curdir, *, dt=None, snaps_from='parallel', read_mode=2):
    '''returns (snapnames, times).

    Gets snapnames from files in 'parallel' directory.
        E.g. parallel000073.h5 --> snapname '73'.
    If no 'parallel' directory, returns ([], []).

    dir: string (default '.')
        directory of 'eppic.i' file.
    dt: None or value
        the simulation time step; each snap's time is equal to dt * (snap name)
        if None, attempt to read 'dt' from eppic.i file.
    snaps_from: 'parallel' or 'timers'
        where to get snapnames from.
        parallel --> get from parallel directory.
        timers --> get from timers.dat file.
    '''
    if snaps_from == 'parallel':
        with InDir(dir):  # temporarily cd to dir; restore original dir after.
            # get dt
            if dt is None:
                globs, _dists = read_eppic_i_file()
                dt = globs['dt']
            # get snaps
            snapnums = []
            if not os.path.isdir('parallel'):
                return ([], [])
            for filename in os.listdir('parallel'):  # directory named 'parallel'
                if read_mode == 2 or read_mode == 3:
                    match = re.match(r'parallel(\d+)\.h5', filename)
                elif read_mode == 4:
                    match = re.match(r'parallel(\d+)\.bp', filename)
                if match:
                    snapname, = match.groups()
                    snapnums.append(int(snapname))
    elif snaps_from == 'timers':
        timers = read_timers_dat(dir, fix_snaps=False)  # don't fix snaps i.e. don't get snaps from parallel.
        snapnums = timers['it'].data
    else:
        raise InputError(f'Expected snaps_from="parallel" or "timers", but got {snaps_from!r}.')
    snapnums = sorted(snapnums)
    snapnames, times = [], []
    for snapnum in snapnums:
        snapnames.append(str(snapnum))
        times.append(dt * snapnum)
    return (snapnames, times)


''' --------------------- read moments.out files --------------------- '''

def read_moments_out_files(dir=os.curdir):
    """
    Read moments*.out files.
    dir: string (default '.')
        directory of 'eppic.i' file.
        momentsN.out files should be located inside dir/domain000
            or might be dir/domain0000, for "newer" versions of EPPIC.

    returns a 3D dictionary {dist: {moment: {time: value}}}
    """

    dists = {}
    moments_dir = os.path.join(dir, 'domain0000')
    if not os.path.isdir(moments_dir):
        moments_dir = os.path.join(dir, 'domain000')
        if not os.path.isdir(moments_dir):
            raise FileNotFoundError(f'{moments_dir!r} (or similar but ending with domain0000).')

    # GET FILE NAMES AND CORRESPONDING DISTRIBUTION NUMBERS
    with InDir(moments_dir): # temporarily cd to dir; restore original dir after.
        # make a dictionary of moments*.out files with *'s as keys i.e. {0: "moments0.out", 1: "moments1.out", etc.}
        moments_out_files = {}
        for filename in sorted(os.listdir()):
            match = re.fullmatch(r'moments(\d+)\.out',filename)
            if match:
                dist_num, = match.groups()
                moments_out_files[dist_num] = match.string

        # SPECIFY NEW COLUMN NAMES
        column_names = ['Snap', 'Vx_Mean', 'Vx_Variance', 'Vx_Moment3', 'Vx_Moment4', 'Vy_Mean', 'Vy_Variance',
                        'Vy_Moment3', 'Vy_Moment4', 'Vz_Mean', 'Vz_Variance', 'Vz_Moment3', 'Vz_Moment4']

        # LOOP THROUGH FILES
        for distribution_number in moments_out_files:
            filename = moments_out_files[distribution_number]
            # LOAD DATA (WITH NEW COLUMN NAMES)
            data = pd.read_csv(filename, sep=r'\s+', skiprows=1, names=column_names, index_col=0)
            # CONVERT DATA TO DICTIONARY
            data_dict = data.to_dict()
            # APPEND DICTIONARY TO DICTIONARY
            dists[distribution_number] = data_dict

    return dists


''' --------------------- get number of processors used --------------------- '''

def _eppic_mpi_processors_from_jobfile(f):
    '''return number of mpi processors used to run eppic based on this jobfile.
    f: str, path to file containing a line like: "MPI STARTING N PROCESSORS" with N an integer.
    '''
    with open(f, 'r') as file:
        lines = file.readlines()
    lines = [line.strip() for line in lines]
    pattern = re.compile(r'MPI STARTING (\d+) PROCESSORS')
    matches = [pattern.fullmatch(line) for line in lines]
    matches = [match for match in matches if match is not None]
    Ns = [int(match.group(1)) for match in matches]
    if len(Ns) == 0:
        errmsg = f'failed to find any lines like "MPI STARTING N PROCESSORS" in file {f!r}'
        raise FileContentsMissingError(errmsg)
    if len(set(Ns)) > 1:
        errmsg = f'found multiple lines like "MPI STARTING N PROCESSORS" in file {f!r}\ngiving different N={Ns}'
        raise FileContentsConflictError(errmsg)
    return Ns[0]

def n_mpi_processors(dir=os.curdir):
    '''return number of mpi processors used to run the run at this directory.
    searches all jobfiles (see find_jobfiles) to determine number of processors,
        from line like "MPI STARTING N PROCESSORS" with N an integer.
    if result is ambiguous (different N from multiple places), raise a FileContentsConflictError.
    if no jobfiles found, raise FileNotFoundError.
    if some jobfiles found but none tell us how many processors were used, raise FileContentsMissingError.
    '''
    dir = os.path.abspath(dir)
    jobfiles = find_jobfiles(dir)
    if len(jobfiles) == 0:
        raise FileNotFoundError(f'no jobfiles found in directory {dir!r}')
    Ns = []
    for f in jobfiles:
        try:
            N = _eppic_mpi_processors_from_jobfile(f)
        except FileContentsMissingError:
            continue  # skip this jobfile, doesn't contain any mpi processor info.
        else:
            Ns.append(N)
    if len(Ns) == 0:
        raise FileContentsMissingError(f'no jobfiles found in directory {dir!r} which contain mpi processor info.')
    if len(set(Ns)) > 1:   # -- note, the len=0 case is handled by _eppic_mpi_processors_from_jobfile.
        errmsg = (f'found multiple lines like "MPI STARTING N PROCESSORS" in jobfiles {jobfiles!r}\n'
                    f'giving different N={Ns}')
        raise FileContentsConflictError(errmsg)
    return Ns[0]


''' --------------------- get wall clock time from jobfile --------------------- '''

def eppic_clock_times_from_jobfile(f):
    '''return clock times from jobfile.
    result can have keys 'start', 'stepstart', 'end':
        start: datetime telling when the run started. From line like:
            "EPPIC Starting at: Friday 02/28/25 12:32:47"
        stepstart: datetime telling when the iterations started. From line like:
            "; --- iterations starting at time: Friday 02/28/25 12:33:31"
        end: datetime telling when the run ended. Various possibilities.
            (1) run ended normally. Two options:
                (a) (this one is tested first): eppic message. Line looks like:
                    "EPPIC ending normally at: Thursday 02/27/25 13:38:20
                (b): slurm file printout. Line looks like:
                    "job total time: NNN seconds"
            (2) timeout due to hitting time limit. Line like ('...' can be any characters):
                "slurmstepd: error: ... CANCELLED AT 2025-02-28T13:32:14 DUE TO TIME LIMIT ***"
            (3) eppic crashed. Line looks like:
                "job total time: NNN seconds"

    result can also have keys 'init_seconds', 'steps_seconds', 'total_seconds':
        init_seconds: (stepstart - start) [seconds]
        steps_seconds: (end - stepstart), [seconds]
        total_seconds: (end - start), [seconds]
    '''
    result = {}
    with open(f, 'r') as file:
        lines = file.readlines()
    lines = [line.strip() for line in lines]
    # start
    pattern = r'EPPIC Starting at: (.+)'
    matches = [re.fullmatch(pattern, line) for line in lines]
    matches = [match for match in matches if match is not None]
    if len(matches) > 1:
        raise FileContentsError(f'multiple lines like {pattern!r} in file {f!r}')
    elif len(matches) == 1:
        tt, = matches[0].groups()
        result['start'] = datetime.datetime.strptime(tt, '%A %m/%d/%y %H:%M:%S')
    # stepstart
    pattern = r'; --- iterations starting at time: (.+)'
    matches = [re.fullmatch(pattern, line) for line in lines]
    matches = [match for match in matches if match is not None]
    if len(matches) > 1:
        raise FileContentsError(f'multiple lines like {pattern!r} in file {f!r}')
    elif len(matches) == 1:
        tt, = matches[0].groups()
        result['stepstart'] = datetime.datetime.strptime(tt, '%A %m/%d/%y %H:%M:%S')
    # end (1)(a)
    pattern = 'EPPIC ending normally at: (.+)'
    matches = [re.fullmatch(pattern, line) for line in lines]
    matches = [match for match in matches if match is not None]
    if len(matches) > 1:
        raise FileContentsError(f'multiple lines like {pattern!r} in file {f!r}')
    elif len(matches) == 1:
        tt, = matches[0].groups()
        result['end'] = datetime.datetime.strptime(tt, '%A %m/%d/%y %H:%M:%S')
    # end (2)
    if 'end' not in result:
        pattern = r'(?:.*)CANCELLED AT (.+) DUE TO TIME LIMIT(?:.*)'
        matches = [re.fullmatch(pattern, line) for line in lines]
        matches = [match for match in matches if match is not None]
        if len(matches) > 1:
            raise FileContentsError(f'multiple lines like {pattern!r} in file {f!r}')
        elif len(matches) == 1:
            tt, = matches[0].groups()
            result['end'] = datetime.datetime.strptime(tt, '%Y-%m-%dT%H:%M:%S')
    # end (3), and (1)(b)
    if 'end' not in result:
        pattern = r'job total time: (\d+) seconds'
        matches = [re.fullmatch(pattern, line) for line in lines]
        matches = [match for match in matches if match is not None]
        if len(matches) > 1:
            raise FileContentsError(f'multiple lines like {pattern!r} in file {f!r}')
        elif len(matches) == 1:
            tt, = matches[0].groups()
            result['end'] = result['start'] + datetime.timedelta(seconds=int(tt))
    # durations
    if 'start' in result and 'stepstart' in result:
        result['init_seconds'] = (result['stepstart'] - result['start']).total_seconds()
    if 'stepstart' in result and 'end' in result:
        result['steps_seconds'] = (result['end'] - result['stepstart']).total_seconds()
    if 'start' in result and 'end' in result:
        result['total_seconds'] = (result['end'] - result['start']).total_seconds()
    return result

def eppic_clock_times_here(dir=os.curdir):
    '''return clock times from jobfile(s) within this directory.
    Crash if multiple jobfiles within this directory tell conflicting times.

    result can have keys 'start', 'stepstart', 'end':
        start: datetime telling when the run started.
        stepstart: datetime telling when the iterations started.
        end: datetime telling when the run ended.

    result can also have keys 'init_seconds', 'steps_seconds', 'total_seconds':
        init_seconds: (stepstart - start) [seconds]
        steps_seconds: (end - stepstart), [seconds]
        total_seconds: (end - start), [seconds]
    '''
    dir = os.path.abspath(dir)
    jobfiles = find_jobfiles(dir)
    if len(jobfiles) == 0:
        raise FileNotFoundError(f'no jobfiles found in directory {dir!r}')
    results = []
    for f in jobfiles:
        results.append(eppic_clock_times_from_jobfile(f))
    # check for conflicts, and combine results into one place
    result = {}
    allkeys = ['start', 'stepstart', 'end', 'init_seconds', 'steps_seconds', 'total_seconds']
    for key in allkeys:
        values = [r[key] for r in results if key in r]
        if len(values) > 1:
            if any(v != values[0] for v in values[1:]):
                errmsg = (f'conflicting values for {key!r} from jobfiles {jobfiles!r}:\n'
                            f'{values}')
                raise FileContentsConflictError(errmsg)
        if len(values) > 0:
            result[key] = values[0]
    return result


''' --------------------- get timers.dat info --------------------- '''

def read_timers_dat(dir=os.curdir, filename='domain0000/timers.dat', *, fix_snaps=True, as_array=False):
    '''read timers.dat info into an xarray.Dataset.
    The "iter #" will be used as the time coordinate, but renamed to "it".
    
    filename: str
        path to timers.dat file.
        if can't find it, but it has 'domain000' or 'domain0000', check the other one.
            E.g. 'domain0000/timers.dat' doesn't exist --> check if 'domain000/timers.dat' exists.
    fix_snaps: bool
        whether to fill NaNs if the last snap is missing from timers.dat, but in parallel folder,
        which can occur if run crashes after saving snap but before writing to timers.dat.
    as_array: bool
        whether to use xarray.Dataset.to_array() to return a DataArray instead of Dataset.
        if True, vars from Dataset will be concatenated along the new dimension named 'timer'.
    '''
    filename = os.path.abspath(os.path.join(dir, filename))
    if not os.path.exists(filename):
        filename_orig = filename
        if 'domain0000' in filename_orig:
            filename = filename_orig.replace('domain0000', 'domain000')
        elif 'domain000' in filename_orig:
            filename = filename_orig.replace('domain000', 'domain0000')
        if not os.path.exists(filename):
            raise FileNotFoundError(f'neither exists: {filename_orig!r},\n{filename!r}')
    ## determine any lines to skip
    # skip any reprinted header (due to continuing run from rst files).
    with open(filename, 'r') as file:
        lines = file.readlines()
    lines = [line.lstrip() for line in lines]
    its, otherstuff = zip(*[line.split(maxsplit=1) for line in lines])   # 'iter #', remaining part of line.
    iheaders = [i for i, it in enumerate(its) if not it.isdigit()]
    # skip any repeated its (due to continuing run from rst files).
    #   (e.g., it 0, 100, 200, save rst, 300, stop. restart --> it 300, 400, ....
    #    Note that it 300 appeared twice. Keep only the last appearance of it,
    #    since the last appearance corresponds to values actually saved to a snapshot.)
    it_counts = collections.Counter(its)
    irepeats = []
    for i, it in enumerate(its):
        if it_counts[it] > 1:
            it_counts[it] -= 1
            irepeats.append(i)
        else:
            del it_counts[it]
    # all lines to skip:
    iskip = set(iheaders).union(set(irepeats)) - set([0])  # don't skip row 0 which has the header.
    ## pandas dataframe from file data:
    # use sep=\s\s+ because separation between cols is at least 2 spaces,
    #   while some column header names contain single spaces in them.
    df = pd.read_csv(filename, sep=r'\s\s+', engine='python', skiprows=iskip)  # engine='python' to use regex matching.
    ## fix snaps (if any snap got saved but missing from timers.dat)
    if fix_snaps:
        snapnames, _times = read_eppic_snaps_info(dir)
        isnaps = [int(snapname) for snapname in snapnames]
        if len(df) != len(snapnames):
            if len(df) == len(snapnames) - 1:  # last snap didn't get written to timers.dat.
                df.loc[len(df), 'iter #'] = isnaps[-1]  # new row of NaN values with the correct snap number.
    # convert to xarray.Dataset
    result = xr.Dataset.from_dataframe(df)
    result = result.assign_coords({'it': ('index', result['iter #'].data)})  
    result = result.swap_dims({'index': 'it'}).drop_vars(['index', 'iter #'])
    # if as_array, convert to xarray.DataArray
    if as_array:
        result = result.to_array(dim='timer')
    return result



"""
File Purpose: Dimension values & lists for Ebysus
"""

from ...dimensions import Fluid, FluidList


''' --------------------- EbysusFluid and EbysusFluidList --------------------- '''

class EbysusFluid(Fluid):
    '''a single Fluid for Ebysus. Similar to Fluid but also knows about SL.
    SL tells (ispecie, ilevel). SL will always internally be stored as tuple(SL).
    '''
    _kw_def = {'name', 'i', 'm', 'q', 'SL'}

    def __init__(self, name=None, i=None, *, m=None, q=None, SL=None):
        super().__init__(name=name, i=i, m=m, q=q)
        self.SL = tuple(SL)  # storing as tuple prevents crash if SL is array,
        # if checking whether SL here == other SL. (Numpy requires using np.all(SL0==SL1)).

    def _repr_contents(self):
        '''contents used by self.__repr__'''
        result = super()._repr_contents()
        if self.SL is not None:
            result.append(f'SL={self.SL}')
        return result


class EbysusFluidList(FluidList):
    '''a list of EbysusFluid objects.'''
    value_type = EbysusFluid

    # pass... all other behavior is inherited from FluidList.



"""
Package Purpose: addons related to analysis of instabilities in general
"""

from .instability_data_tools import (
    pwl2_flatend, pwl3_flatend,
    Pwl2FlatendFitter, Pwl3FlatendFitter,
)
from .instability_quantity_loader import InstabilityQuantityLoader
from .instability_theory_tools import (
    xarray_khat, xarray_k,
    xarray_at_growmax,
    xarray_growth_kmax, xarray_grows,
    xarray_kmod_at_growmax, xarray_kang_at_growmax,
    xarray_khat_at_growmax, xarray_kds_at_growmax, xarray_k_at_growmax,
    xarray_smod_vphase, xarray_mod_vphase, xarray_vphase,
    xarray_where_grows, xarray_where_nogrows,
    xarray_stack_nonk_dims,
    xarray_kw_growthplot, xarray_growthplots, xarray_growthplot,
    xarray_klims_physical, xarray_klines,
)
from .instability_xarray_accessor import (
    itAccessor, itArrayAccessor, itDatasetAccessor,
)


"""
File Purpose: testing array-related methods
"""
import pytest

import PlasmaCalcs as pc

def test_fractional_indexing():
    '''tests that interprets_fractional_indexing works as intended.'''
    # leave ints unchanged
    ints = [0, 1, 2, 3, 4, 5, -1, -2, -3, -4, -5]
    for i in ints:
        assert i == pc.interprets_fractional_indexing(i)
        assert slice(i) == pc.interprets_fractional_indexing(slice(i))
        assert slice(None, i) == pc.interprets_fractional_indexing(slice(None, i))
        assert slice(None, None, i) == pc.interprets_fractional_indexing(slice(None, None, i))
    # handle list
    assert ints == pc.interprets_fractional_indexing(ints)
    # handle fractions when L is provided
    ## 0.5 --> halfway through the list
    ### for odd lengths, that is the unique middle index (regardless of rounding)
    odds = [3, 5, 7, 9, 27, 123, 999]
    for L in odds:
        I = (L-1)//2
        assert I == pc.interprets_fractional_indexing(0.5, L=L)
        for rounding in ('round', 'int', 'floor', 'ceil'):
            assert I == pc.interprets_fractional_indexing(0.5, L=L, rounding=rounding)
        ## also check that -0.5 works as expected:
        assert -I == pc.interprets_fractional_indexing(-0.5, L=L)
        for rounding in ('round', 'int', 'floor', 'ceil'):
            assert -I == pc.interprets_fractional_indexing(-0.5, L=L, rounding=rounding)
        ## also check that slices & iterables work as expected:
        assert slice(I, None, None) == pc.interprets_fractional_indexing(slice(0.5, None, None), L=L)
        assert slice(None, I, None) == pc.interprets_fractional_indexing(slice(None, 0.5, None), L=L)
        assert slice(None, None, I) == pc.interprets_fractional_indexing(slice(None, None, 0.5), L=L)
        assert slice(-7, 200, I) == pc.interprets_fractional_indexing(slice(-7, 200, 0.5), L=L)
        assert [0, -1, I, 2, -2, 1, I] == pc.interprets_fractional_indexing([0, -1, 0.5, 2, -2, 1, 0.5], L=L)
    ### for even lengths, 0.5 depends on rounding.
    evens = [4, 6, 8, 10, 28, 124, 1000]
    for L in evens:
        Ifloor = L//2 - 1
        Iceil = L//2
        assert Ifloor == pc.interprets_fractional_indexing(0.5, L=L, rounding='floor')
        assert Iceil == pc.interprets_fractional_indexing(0.5, L=L, rounding='ceil')
        ## also check that -0.5 works as expected (abs for floor & ceil swap below 0):
        assert -Ifloor == pc.interprets_fractional_indexing(-0.5, L=L, rounding='ceil')
        assert -Iceil == pc.interprets_fractional_indexing(-0.5, L=L, rounding='floor')
    # handle exceptions / edge cases
    inputs = [0.2, -0.7,
            slice(0.2, 10), slice(-0.7, 10), slice(-100, 0.3),
            [0.2, 0.3, -0.7, 0.5], [100, -8, 0.3, 9, 10]]
    ## crash when L not provided but fractional indexing is required:
    for i in inputs:
        with pytest.raises(pc.InputMissingError):
            pc.interprets_fractional_indexing(i)  # didn't provide L but fractional indexing required.
    ## crash when L <= 1 but only if fractional indexing is required:
    for i in inputs:
        with pytest.raises(ValueError):
            pc.interprets_fractional_indexing(i, L=0)
        with pytest.raises(ValueError):
            pc.interprets_fractional_indexing(i, L=1)
        with pytest.raises(ValueError):
            pc.interprets_fractional_indexing(i, L=-20)
    assert pc.interprets_fractional_indexing(7, L=0) == 7   # no crash; fractional indexing not required.
    assert pc.interprets_fractional_indexing(slice(0, 7, 3), L=1) == slice(0, 7, 3)
    assert pc.interprets_fractional_indexing([9, 8, 2, -7, 5], L=-20) == [9, 8, 2, -7, 5]



"""
File Purpose: test vector arithmetic patterns from quantity loader.
"""
import PlasmaCalcs as pc

import numpy as np


def test_vector_arithmetic_basics():
    '''tests basics of vector arithmetic.'''
    p = pc.PlasmaCalculator()  # plasma calculator object not associated with any files.
    # components dimension
    assert len(p.components) == 3
    p.component = None  # use all components
    assert len(p.component) == 3
    assert p.components == pc.XYZ
    assert p.components == ['x', 'y', 'z']
    assert p.components == [0, 1, 2]
    assert p.components == ['x', 1, pc.XYZ[2]]
    # unit vectors & components
    assert np.all(p('xhat') == pc.XHAT)
    assert np.all(p('yhat') == pc.YHAT)
    assert np.all(p('zhat') == pc.ZHAT)
    for (x, y, z) in ('xyz', 'yzx', 'zxy'):
        assert p(f'{x}hat_{x}').item() == 1   # {x}hat == 1 in the {x} direction
        assert p(f'{x}hat_{y}').item() == 0   # {x}hat == 0 in the {y} direction
        assert p(f'{x}hat_{z}').item() == 0   # {x}hat == 0 in the {z} direction
    # composing vectors from components
    assert np.all(p('-xhat') == -p('xhat'))
    assert np.all(p('-xhat') == [-1, 0, 0])
    assert np.all(p('xhat*7') == [7, 0, 0])
    assert np.all(p('3*yhat+2*zhat') == [0, 3, 2])
    assert np.all(p('xhat*7+yhat*3+zhat*2') == [7, 3, 2])
    assert np.all(p('-xhat*7+3*yhat-zhat*2') == [-7, 3, -2])
    # dot products of pairs of unit vectors
    for x in 'xyz':
        for y in 'xyz':
            assert p(f'{x}hat_dot_{y}hat').item() == (x == y)
    # dot products of vectors composed of unit vectors
    assert p('(3*xhat)_dot_xhat').item() == 3
    assert p('(3*xhat+2*yhat)_dot_xhat') == 3
    assert p('(3*xhat+2*yhat)_dot_yhat') == 2
    assert p('(3*xhat+2*yhat)_dot_zhat') == 0
    assert p('(3*xhat+yhat+4*zhat)_dot_(7*xhat+50*zhat)') == 7*3 + 50*4
    # cross products of pairs of unit vectors
    assert np.all(p('xhat_cross_yhat') == p('zhat'))
    assert np.all(p('yhat_cross_zhat') == p('xhat'))
    assert np.all(p('zhat_cross_xhat') == p('yhat'))
    assert np.all(p('yhat_cross_xhat') == p('-zhat'))
    assert np.all(p('zhat_cross_yhat') == p('-xhat'))
    assert np.all(p('xhat_cross_zhat') == p('-yhat'))
    # cross products of vectors composed of unit vectors
    assert np.all(p('(3*xhat)_cross_xhat') == 0)
    assert np.all(p('(3*xhat+2*yhat)_cross_xhat') == [0, 0, -2])
    assert np.all(p('(3*xhat+2*yhat)_cross_xhat') == p('-2*zhat'))
    assert np.all(p('(3*xhat+2*yhat)_cross_yhat') == p('3*zhat'))
    assert np.all(p('(3*xhat+2*yhat)_cross_yhat') == [0, 0, 3])
    assert np.all(p('(3*xhat+2*yhat)_cross_zhat') == p('2*xhat-3*yhat'))
    assert np.all(p('(3*xhat+2*yhat)_cross_zhat') == p('-3*yhat+2*xhat'))
    assert np.all(p('(3*xhat+2*yhat)_cross_zhat') == [2, -3, 0])
    assert np.all(p('(3*xhat+yhat+4*zhat)_cross_(7*xhat+50*zhat)') == [1*50, -3*50 + 7*4, -1*7])  # [50, -122, -7]
    # using cross product & dot product functions directly:
    for xhat in (pc.XHAT, pc.YHAT, pc.ZHAT):
        for yhat in (pc.XHAT, pc.YHAT, pc.ZHAT):
            assert np.all(pc.dot_product(xhat, yhat).item() == (xhat is yhat))
    assert np.all(pc.cross_product(pc.XHAT, pc.YHAT) == pc.ZHAT)
    assert np.all(pc.cross_product(pc.YHAT, pc.ZHAT) == pc.XHAT)
    assert np.all(pc.cross_product(pc.ZHAT, pc.XHAT) == pc.YHAT)
    assert np.all(pc.cross_product(pc.YHAT, pc.XHAT) == -pc.ZHAT)
    assert np.all(pc.cross_product(pc.ZHAT, pc.YHAT) == -pc.XHAT)
    assert np.all(pc.cross_product(pc.XHAT, pc.ZHAT) == -pc.YHAT)



"""
File Purpose: testing some parts of plotting code.
"""
import os

import matplotlib.pyplot as plt
import numpy as np
import pytest
import xarray as xr

import PlasmaCalcs as pc

# ARTIFACTS STUFF:
HERE = os.path.dirname(__file__)
TESTS_DIR = HERE
ARTIFACTS = os.path.join(TESTS_DIR, '_test_artifacts', 'plots')
if os.path.exists(ARTIFACTS):  # remove any plots artifacts before running tests.
    import shutil; shutil.rmtree(ARTIFACTS)
os.makedirs(ARTIFACTS)


def savefig_and_close(name, *, bbox_inches='tight', **kw):
    '''savefig to ARTIFACTS/name, then plt.close().'''
    plt.savefig(os.path.join(ARTIFACTS, name), bbox_inches=bbox_inches, **kw)
    plt.close()

def test_plot_settings_cls():
    '''testing the PlotSettings class'''
    keys = list(pc.PlotSettings.MPL_KWARGS.keys())
    assert len(keys) > 0   # i.e., this test isn't pointless...
    assert len(keys) > 5   # there are more than 5 keys at time of writing this test.
    for key in keys:
        vals = pc.PlotSettings.get_mpl_keys(key)  # ensure this doesn't crash
        assert len(vals) > 0         # and this key points to at least 1 kwarg.

    ps = pc.PlotSettings(squeeze=True, figsize=(10, 10), dpi=100, fontsize=12, aspect=1)
    kw = ps.get_mpl_kwargs('fig.subplots')
    assert kw['squeeze'] is True
    assert kw['subplot_kw']['aspect'] == 1
    assert all(key not in kw for key in ('dpi', 'figsize', 'fontsize')) # these cannot be passed to fig.subplots.
    # note, kw might contain other values if defaults have changed since this test was written.
    kw = ps.get_mpl_kwargs('plt.figure')
    default_layout = ps.get_default('layout')  # layout is one of the figure kwargs with a default value.
    assert kw['layout'] == default_layout
    assert kw['figsize'] == (10, 10)
    assert kw['dpi'] == 100
    assert all(key not in kw for key in ('squeeze', 'fontsize')) # these cannot be passed to plt.figure.

def test_plot_locations():
    '''tests that plot_locations() & plot_note() works! Also tests ax_remove_ticks().
    Also produce artifact that can be checked by eye later & used as a reference for locations.
    '''
    locations = pc.plot_locations()   # dict of {loc: {'xy': (x,y), 'ha': str, 'va': str}}
    for loc in locations:
        # use 'here:\n' to make multiline comment to confirm proper va.
        # don't need to test locations[loc] values because plot_note also uses plot_locations().
        pc.plot_note('here:\n'+loc, loc)
    pc.ax_remove_ticks()
    savefig_and_close('test_plot_locations.png')

def test_xarray_timelines():
    '''test that xarray timelines doesn't crash...
    also produce artifacts that could be checked by eye later.
    '''
    pc.plot_note('''test_xarray_timelines plot00, check:
        4 lines with 7 points each; slope=4;
        offset = 0, 1, 2, 3; t = 0, 1, 2, 3, 4, 5, 6;
        colors vary with offset; markers are 'o'.''')
    arr00 = xr.DataArray(np.arange(7*4).reshape(7, 4), dims=['t', 'offset'])
    tls00 = arr00.pc.timelines(marker='o')
    savefig_and_close('test_xarray_timelines_00.png')
    # plot 01: like plot 00 but all lines are blue & markers vary.
    pc.plot_note('''test_xarray_timelines plot01, check:
        4 lines with 7 points each; slope=4;
        markers vary with offset; all lines are blue.''')
    tls01 = arr00.pc.timelines(cycles=[dict(marker=['o', 'x', '+', 's'])], color='blue')
    savefig_and_close('test_xarray_timelines_01.png')
    # plot 02: like plot 00 but t varies from 0 to 1. Also ensure adding misc coord is fine!
    pc.plot_note('''test_xarray_timelines plot02, check:
        4 lines with 7 points each; t varies from 0 to 1.''')
    arr02 = xr.DataArray(np.arange(7*4).reshape(7, 4), dims=['t', 'offset'],
                        coords=dict(t=np.linspace(0, 1, 7), misc_coord='wow'))
    tls02 = pc.XarrayTimelines(arr02, marker='o')  # ensure pc.XarrayTimelines(arr) syntax works too!
    assert np.all(pc.get_data_interval('x') == np.array([0,1]))  # ensure x-axis is 0 to 1.
    savefig_and_close('test_xarray_timelines_02.png')
    # plot 10: similar idea to plot00 but introduce a third dimension (shift)
    pc.plot_note('''test_xarray_timelines plot10, check:
        12 lines with 7 points each; slope=4;
        colors vary with offset;
        linestyles vary with shift.''')
    arr10 = xr.DataArray(np.arange(7*4*3).reshape(7, 4, 3), dims=['t', 'offset', 'shift'])
    tls10 = arr10.pc.timelines()
    savefig_and_close('test_xarray_timelines_10.png')

def test_xarray_image_gif():
    '''test that XarrayImage doesn't crash when making gif...
    also produce artifacts that could be checked by eye later.
    '''
    pc.plot_note('''test_xarray_image_gif plot00, check:
        10x20 image, 8 frames; t varies 0 to 2*pi;
        x varies 0 to 2*pi; y varies 0 to 4*pi;
        waves go lower left (0,0) to upper right;
        final frame (t=2*pi) = first frame (t=0).''',
        'outside upper left', margin=0.15, fontsize='small')
    coords = {'x': np.linspace(0, 2*np.pi, 10),
              'y': np.linspace(0, 4*np.pi, 20),
              't': np.linspace(0, 2*np.pi, 8)}
    xx = xr.DataArray(coords['x'], coords=dict(x=coords['x']))
    yy = xr.DataArray(coords['y'], coords=dict(y=coords['y']))
    tt = xr.DataArray(coords['t'], coords=dict(t=coords['t']))
    data = np.sin(xx + yy - tt)
    xim = data.pc.image(title='t={t:.2f}')
    xim.save(os.path.join(ARTIFACTS, 'test_xarray_image_gif_00.gif'), fps=10)

# [TODO] how to tell gitlab to render mp4?
# def test_xarray_image_mp4():
#     '''test that XarrayImage doesn't crash when making mp4...
#     also produce artifacts that could be checked by eye later.

#     mp4 separate from gif in case one of them crashes
#     '''
#     pc.plot_note('''test_xarray_image_mp4 plot00, check:
#         10x20 image, 8 frames; t varies 0 to 2*pi;
#         x varies 0 to 2*pi; y varies 0 to 4*pi;
#         waves go lower left (0,0) to upper right;
#         final frame (t=2*pi) = first frame (t=0).''',
#         'outside upper left', margin=0.15, fontsize='small')
#     coords = {'x': np.linspace(0, 2*np.pi, 10),
#               'y': np.linspace(0, 4*np.pi, 20),
#               't': np.linspace(0, 2*np.pi, 8)}
#     xx = xr.DataArray(coords['x'], coords=dict(x=coords['x']))
#     yy = xr.DataArray(coords['y'], coords=dict(y=coords['y']))
#     tt = xr.DataArray(coords['t'], coords=dict(t=coords['t']))
#     data = np.sin(xx + yy - tt)
#     xim = data.pc.image(title='t={t:.2f}')
#     xim.save(os.path.join(ARTIFACTS, 'test_xarray_image_mp4_00.mp4'), fps=10)

def test_faceplot():
    '''test that Faceplot doesn't crash when making gif...
    also produce artifacts that could be checked by eye later.
    '''
    # create "dataset":
    coords = {'x': np.linspace(0, 4*np.pi, 20),
          'y': np.linspace(0, 2*np.pi, 10),
          'z': np.linspace(0, 2*np.pi, 10),
          't': np.linspace(0, 2*np.pi, 8)}
    xx = xr.DataArray(coords['x'], coords=dict(x=coords['x']))
    yy = xr.DataArray(coords['y'], coords=dict(y=coords['y']))
    zz = xr.DataArray(coords['z'], coords=dict(z=coords['z']))
    tt = xr.DataArray(coords['t'], coords=dict(t=coords['t']))
    data = dict()
    data['x_y'] = np.sin(xx + yy + 0 - tt)
    data['x_z'] = np.sin(xx + 0 + zz - tt)
    data['y_z'] = np.sin(0 + yy + zz - tt)
    data = xr.Dataset(data)
    # plot00: test that faceplot works
    fp = data.pc.faceplot(title='t={t:.2f}')
    pc.plot_note('''test_faceplot plot00, check:
        x varies 0 to 4*pi;  y,z,t vary 0 to 2*pi;
        waves move away from origin (0,0,0);
        final frame (t=2*pi) = first frame (t=0).''',
        'upper center', fontsize='small')
    fp.save(os.path.join(ARTIFACTS, 'test_faceplot_00.gif'), fps=10)
    # plot01: should produce same result as plot00, but the input is full 3D data;
    #   ensures that "infer faces from 3D data" works properly.
    data = np.sin(xx + yy + zz - tt)
    fp = data.pc.faceplot(title='t={t:.2f}')
    pc.plot_note('''test_faceplot plot01, check:
        x varies 0 to 4*pi;  y,z,t vary 0 to 2*pi;
        waves move away from origin (0,0,0);
        final frame (t=2*pi) = first frame (t=0).''',
        'upper center', fontsize='small')
    fp.save(os.path.join(ARTIFACTS, 'test_faceplot_01.gif'), fps=10)

def test_xarray_line():
    '''test that XarrayLine & XarrayLines don't crash...
    also produce artifacts that could be checked by eye later.
    '''
    # plot00: test that XarrayLine works
    pc.plot_note('''test_xarray_line plot00, check:
        x varies 0 to 4*pi; t varies 0 to 2*pi;
        waves move left to right; marker='o';
        final frame (t=2*pi) = first frame (t=0).''',
        fontsize='small', fontweight='bold')
    xx = pc.xr1d(np.linspace(0, 4*np.pi, 20), 'x')
    tt = pc.xr1d(np.linspace(0, 2*np.pi, 8), 't')
    data = np.sin(xx - tt)
    xline = data.pc.line(title='t={t:.2f}', marker='o')
    xline.save(os.path.join(ARTIFACTS, 'test_xarray_line_00.gif'), fps=10)
    # plot10: test that XarrayLines works.
    pc.plot_note('''test_xarray_line plot10, check:
        x varies 0 to 4*pi; t varies 0 to 2*pi;
        waves move left to right;
        final frame (t=2*pi) = first frame (t=0);
        3 similar waves offset by 2, 0, -1;
        legend tells offset for each line.''',
        fontsize='small', fontweight='bold')
    xx = pc.xr1d(np.linspace(0, 4*np.pi, 20), 'x')
    tt = pc.xr1d(np.linspace(0, 2*np.pi, 8), 't')
    oo = pc.xr1d(np.array([2, 0, -1]), 'offset')
    data = oo + np.sin(xx - tt)
    xlines = data.pc.lines(title='t={t:.2f}') # label='offset={offset}'
    xlines.save(os.path.join(ARTIFACTS, 'test_xarray_line_10.gif'), fps=10)
    # ensure that we raise DimensionError if too many dimensions for XarrayLine (singular)
    with pytest.raises(pc.DimensionError):
        data.pc.line()  # singular! But data has 3 dims --> wrong shape for only 1 line.

def test_xarray_rectangle_patch():
    '''test that XarrayRectanglePatch doesn't crash...
    also produce artifacts that could be checked by eye later.
    '''
    # plot00: test that XarrayRectanlgePatch works
    pc.plot_note('''test_xarray_rectangle_patch plot00, check:
        rectangle anchored at 7th point along x;
        rectangle wider but less tall over time.''',
        fontsize='small', fontweight='bold')
    xx = pc.xr1d(np.linspace(0, 4*np.pi, 20), 'x')
    tt = pc.xr1d(np.linspace(0, 2*np.pi, 8), 't')
    # line:
    data = np.sin(xx - tt)
    xline = data.pc.line(title='t={t:.2f}', marker='o')
    # rectangle:
    rinfo = xr.Dataset(dict(
        x0 = xx.isel(x=7),
        y0 = data.isel(x=7),
        width = (1+tt),
        height = 0.75 / (1+tt),
    ))
    xrect = rinfo.pc.rectangle_patch(facecolor='none', hatch=r'/\\/', edgecolor='black')
    # saving animation
    xline.add_child(xrect) # xline needs to learn about xrect so it knows to animate it:
    xline.save(os.path.join(ARTIFACTS, 'test_xarray_rectangle_patch_00.gif'), fps=10)


def test_xarray_subplots():
    '''test that XarraySubplots doesn't crash...
    also produce artifacts that could be checked by eye later.
    '''
    xx = pc.xr1d(np.linspace(0, 4 * np.pi, 20), 'x')
    yy = pc.xr1d(np.linspace(0, 2 * np.pi, 10), 'y')
    yshift = pc.xr1d([0, np.pi/2, np.pi], 'yshift')
    offset = pc.xr1d([0, -1], 'offset')
    xshift = pc.xr1d(np.linspace(0, 2 * np.pi, 8), 'xshift')
    # plot00: test that XarraySubplots works (3D data)
    suptitle = '''test_xarray_subplots plot00, check:
        x varies 0 to 4*pi; y varies 0 to 2*pi;
        values = cos(x) + sin(y - yshift);
        3 plots in col across yshift = [0, pi/4, pi/2]'''
    ff = (np.cos(xx) + np.sin(yy - yshift))
    ff.pc.subplots(col='yshift',
                   suptitle=suptitle)
    savefig_and_close('test_xarray_subplots_00.png')
    # plot10: another dimension (4D data)
    suptitle = '''test_xarray_subplots plot10, check:
        x varies 0 to 4*pi; y varies 0 to 2*pi;
        values = cos(x) + sin(y - yshift) + offset;
        3 plots in col across yshift = [0, pi/4, pi/2];
        2 plots in row across offset = [0, -1];
        each plot has its own colorbar.'''
    ff = (np.cos(xx) + np.sin(yy - yshift) + offset)
    ff.pc.subplots(col='yshift', row='offset',
                   suptitle=suptitle)
    savefig_and_close('test_xarray_subplots_10.png')
    # plot11: share vlims
    suptitle = '''test_xarray_subplots plot11, check:
        x varies 0 to 4*pi; y varies 0 to 2*pi;
        values = cos(x) + sin(y - yshift) + offset;
        3 plots in col across yshift = [0, pi/4, pi/2];
        2 plots in row across offset = [0, -1];
        shared colors, and colorbars only in right col.'''
    ff = (np.cos(xx) + np.sin(yy - yshift) + offset)
    ff.pc.subplots(col='yshift', row='offset',
                   suptitle=suptitle, share_vlims=True)
    savefig_and_close('test_xarray_subplots_11.png')
    # plot20: yet another dimension (5D data)
    suptitle = '''test_xarray_subplots plot20, check:
        x varies 0 to 4*pi; y varies 0 to 2*pi;
        values = cos(x-xshift) + sin(y-yshift) + offset;
        3 plots in col across yshift = [0, pi/4, pi/2];
        2 plots in row across offset = [0, -1];
        shared colors, and colorbars only in right col.
        xshift varies in time from 0 to 2*pi,
        and suptitle text updates: xshift={xshift:.2e}'''
    ff = (np.cos(xx - xshift) + np.sin(yy - yshift) + offset)
    xsubs = ff.pc.subplots(col='yshift', row='offset',
                           suptitle=suptitle, share_vlims=True)
    assert xsubs.t_plot_dim == 'xshift'  # t is inferred automatically
    xsubs.save(os.path.join(ARTIFACTS, 'test_xarray_subplots_20.gif'), fps=10)



"""
File Purpose: manage currently active figure, axs, etc.
"""

import matplotlib.pyplot as plt

from ...tools import NO_VALUE

''' --------------------- as functions --------------------- '''

def current_figure_exists():
    '''tells whether there is a "current" figure.'''
    return len(plt.get_fignums()) > 0

def current_figure_or_None():
    '''returns current figure, or None if there is no "current figure".'''
    return plt.gcf() if current_figure_exists() else None

def current_axes_exists():
    '''tells whether there is a "current" axes.'''
    fig = current_figure_or_None()
    return False if fig is None else (len(fig.axes) > 0)

def current_axes_or_None():
    '''returns current axes, or None if there is no "current axes".'''
    return plt.gca() if current_axes_exists() else None

def current_axes_has_data():
    '''returns whether the current axes has any data plotted on it.'''
    ax = current_axes_or_None()
    return False if ax is None else ax.has_data()

def current_image_exists():
    '''tells whether there is a "current" image.'''
    return plt.gci() is not None

def current_image_or_None():
    '''returns current image, or None if there is no "current image".'''
    return plt.gci()


''' --------------------- as context managers --------------------- '''

def maintaining_current_plt(*maintain, default=['all']):
    '''returns context manager which restores current figure, axes, image upon exit.
    If the original figure / axes / image was None, it will not be restored to None. [TODO] change this?
    maintain: strings from ('figure', 'axes', 'image', 'all')
        maintain (only) these values. E.g. ['figure'] --> maintain current figure and axes.
        'all' --> maintain all values.
    default: list of strings, default ['all']
        if len(maintain)==0, instead use maintain = default.

    see also: using_current_plt
    '''
    return MaintainingCurrentPlt(*maintain, default=default)

def using_current_plt(*, figure=NO_VALUE, axes=NO_VALUE, image=NO_VALUE):
    '''returns context manager which sets values for "current" figure, axes, image upon entry; restores upon exit.
    If the original figure / axes / image was None, it will not be restored to None. [TODO] change this?
    figure, axes, image: objects or NO_VALUE
        if provided, set that value as the current value, upon entry.

    see also: maintaining_current_plt
    '''
    return UsingCurrentPlt(figure=figure, axes=axes, image=image)

def maintaining_current_figure(enabled=True):
    '''returns context manager which (if enabled) restores current figure upon exit. See also: maintaining_current_plt'''
    return maintaining_current_plt(*(['figure'] if enabled else []), default=[])
def maintaining_current_axes(enabled=True):
    '''returns context manager which (if enabled) restores current axes upon exit. See also: maintaining_current_plt'''
    return maintaining_current_plt(*(['axes'] if enabled else []), default=[])
def maintaining_current_image(enabled=True):
    '''returns context manager which (if enabled) restores current image upon exit. See also: maintaining_current_plt'''
    return maintaining_current_plt(*(['axes'] if enabled else []), default=[])
def using_current_figure(figure):
    '''returns context manager which sets current figure upon entry; restores upon exit. See also: using_current_plt'''
    return using_current_plt(figure=figure)
def using_current_axes(axes):
    '''returns context manager which sets current axes upon entry; restores upon exit. See also: using_current_plt'''
    return using_current_plt(axes=axes)
def using_current_image(image):
    '''returns context manager which sets current image upon entry; restores upon exit. See also: using_current_plt'''
    return using_current_plt(image=image)

class MaintainingCurrentPlt():
    '''context manager which restores current figure, axes, image upon exit.
    If the original figure / axes / image was None, it will not be restored to None. [TODO] change this?
    maintain: strings from ('figure', 'axes', 'image', 'all')
        maintain (only) these values. E.g. ['figure'] --> maintain current figure and axes.
        'all' --> maintain all values.
    default: list of strings, default ['all']
        if len(maintain)==0, instead use maintain = default.
    '''
    def __init__(self, *maintain, default=['all']):
        if len(maintain)==0:
            maintain = default
        figure = axes = image = False
        if 'all' in maintain:
            figure = axes = image = True
        else:
            figure = ('figure' in maintain)
            axes = ('axes' in maintain)
            image = ('image' in maintain)
        self.maintain_figure = figure
        self.maintain_axes = axes
        self.maintain_image = image

    def __enter__(self):
        self.memory = dict()
        if self.maintain_figure:
            self.memory['figure'] = current_figure_or_None()
        if self.maintain_axes:
            self.memory['axes'] = current_axes_or_None()
        if self.maintain_image:
            self.memory['image'] = current_image_or_None()

    def __exit__(self, *args_unused):
        if self.maintain_figure and (self.memory['figure'] is not None):
            plt.figure(self.memory['figure'])
        if self.maintain_axes and (self.memory['axes'] is not None):
            plt.sca(self.memory['axes'])
        if self.maintain_image and (self.memory['image'] is not None):
            plt.sci(self.memory['image'])


class UsingCurrentPlt(MaintainingCurrentPlt):
    '''context manager which sets values for "current" figure, axes, image upon entry; restores upon exit.
    If the original figure / axes / image was None, it will not be restored to None. [TODO] change this?
    Only actually sets (restores) the values which are (were) provided, via kwargs.
    '''
    def __init__(self, *, figure=NO_VALUE, axes=NO_VALUE, image=NO_VALUE):
        self.figure = figure
        self.axes = axes
        self.image = image
        self.maintain_figure = (figure is not NO_VALUE)
        self.maintain_axes = (axes is not NO_VALUE)
        self.maintain_image = (image is not NO_VALUE)

    def __enter__(self):
        super().__enter__()
        if self.figure is not NO_VALUE:
            plt.figure(self.figure)
        if self.axes is not NO_VALUE:
            plt.sca(self.axes)
        if self.image is not NO_VALUE:
            plt.sci(self.image)



"""
File Purpose: tools for arrays (more generic than xarrays)
"""

import numpy as np

from .docs_tools import format_docstring
from .iterables import is_iterable
from .math import round_to_int
from .sentinels import UNSET
from ..errors import (
    InputError, InputConflictError, InputMissingError, 
    MemorySizeError,
    DimensionalityError,
)
from ..defaults import DEFAULTS


''' --------------------- Size in Memory --------------------- '''

def memory_size_check(array, *, MBmax=UNSET, dtype_min=UNSET, safety=1, errmsg=None):
    '''raise MemorySizeError if array is too large.
    Too large if safety * array.nbytes/1024**2 > MBmax.

    array: array, e.g. numpy array
        must have .size and .dtype attributes which tell number of elements and numpy dtype.
        Or, can be an xarray.Dataset, in which case will add up the size of all DataArrays.
    MBmax: UNSET, None, or int
        maximum size allowed, in Megabytes.
        UNSET --> use DEFAULTS.ARRAY_MBYTES_MAX
        None --> no limit.
    dtype_min: UNSET or numpy dtype
        calculate memory as if array elements will have at least as many bytes as dtype_min.itemsize.
        (if array elements already have more bytes than this, use array.dtype instead.)
        UNSET --> use DEFAULTS.ARRAY_MEMORY_CHECK_DTYPE_MIN
        None --> don't impose any minimum, just use array.dtype.
    safety: number
        safety factor; safety * array.nbytes/1024**2 is the actual limit.
    errmsg: None or str
        if provided, if raising MemorySizeError, use this for the error message,
        after formatting with .format(nMB=nMB, MBmax=MBmax, safety=safety, nMB_total=nMB * safety,
                                    dtype=dtype used, shape=array.shape, size=array.size).
        if None, use:
            "array too large: {nMB:.1e} MB > MBmax / safety. (MBmax={MBmax} MB, safety={safety})".
    '''
    if MBmax is UNSET: MBmax = DEFAULTS.ARRAY_MBYTES_MAX
    if dtype_min is UNSET: dtype_min = DEFAULTS.ARRAY_MEMORY_CHECK_DTYPE_MIN
    if MBmax is not None:
        arrays = [array] if not hasattr(array, 'data_vars') else list(array.data_vars.values())
        # ^ handles xarray.Dataset, which has data_vars which lists arrays.
        nMB = 0
        for array in arrays:
            size = array.size
            dtype = array.dtype
            itemsize = array.dtype.itemsize
            if dtype_min is not None:
                min_itemsize = np.dtype(dtype_min).itemsize
                if min_itemsize > array.dtype.itemsize:
                    dtype = dtype_min
                    itemsize = min_itemsize
            nbytes = size * itemsize
            nMB = nMB + nbytes / 1024**2
        if safety * nMB > MBmax:
            if errmsg is None:
                errmsg = ('array too large: {nMB:.1e} MB > MBmax / safety; '
                          '(MBmax={MBmax} MB, safety={safety})')
            errmsg = errmsg.format(nMB=nMB, MBmax=MBmax, safety=safety, nMB_total=nMB * safety,
                                   dtype=dtype, shape=array.shape, size=array.size)
            raise MemorySizeError(errmsg)

def memory_size_check_loading_arrays_like(array1, nload=1, *, MBmax=UNSET, dtype_min=UNSET):
    '''raise MemorySizeError if loading nload arrays like array1 will be too large.
    Too large if total size in MB would be more than MBmax.

    array1: array, e.g. numpy array
        must have .nbytes attribute, which tells size in bytes.
    n: int
        number of arrays like array1 which will be loaded.
    MBmax: UNSET, None, or int
        maximum size allowed, in Megabytes.
        UNSET --> use DEFAULTS.ARRAY_MBYTES_MAX
        None --> no limit.
    dtype_min: UNSET or numpy dtype
        calculate memory as if array elements will have at least as many bytes as dtype_min.itemsize.
        (if array elements already have more bytes than this, use array.dtype instead.)
        UNSET --> use DEFAULTS.ARRAY_MEMORY_CHECK_DTYPE_MIN
        None --> don't impose any minimum, just use array.dtype.

    return size of array1, in MB.
    '''
    __tracebackhide__ = DEFAULTS.TRACEBACKHIDE
    # this function just calls memory_size_check, with a more helpful error message if it fails.
    errmsg = ("array too large to load n = {safety} simultaneously, with array memory size = {nMB:.4g} MB.\n"
              "Calculated memory with dtype = {dtype}, array.size = {size:.2e} elements, array.shape = {shape}.\n"
              "Consider loading fewer arrays simultaneously. "
              "Alternatively, adjust the limit by setting DEFAULTS.ARRAY_MBYTES_MAX.\n"
              "Current limit = {MBmax:.4g} MB. Here, would need at least {nMB_total:.4g} MB.")
    return memory_size_check(array1, MBmax=MBmax, dtype_min=dtype_min, safety=nload, errmsg=errmsg)


''' --------------------- Finite operations --------------------- '''

def finite_op(arr, op, *, if_empty=UNSET):
    '''returns op(arr), hitting only the finite values of arr.

    arr: array-like or list of differently-sized array-like objects.
        will do op(array, or list of arrays to flatten & concatenate).
        if arr is a single array-like has only finite values,
            finite_op(arr, op) == op(arr).
        if arr is a single array-like and has some nonfinite values (infs or nans),
            finite_op(arr, op) == op(arr[np.isfinite(arr)])
        if np.asanyarray(arr) fails with ValueError,
            attempt to interpret arr as a list of differently-sized array-like objects;
            convert each to array via np.asanyarray, reshape(-1), then concatenate.

    if_empty: UNSET or any value
        value to return if arr is an empty array (or there are no finite values in it).
        UNSET --> use numpy default behavior (e.g. np.mean([]) --> nan; np.min([]) --> crash)
        any value --> return this value.
    '''
    try:
        arr = np.asanyarray(arr)
    except ValueError:  # might be a list of differently-sized arrays-like objects.
        arrs = [np.asanyarray(a).reshape(-1) for a in arr]
        arr = np.concatenate(arrs)
    finite = np.isfinite(arr)
    if np.count_nonzero(finite) < finite.size:
        arr_use = arr[finite]   # [EFF] index by finite only if there's any non-finite values.
    else:
        arr_use = arr
    if (if_empty is not UNSET) and (arr_use.size == 0):
        return if_empty
    return op(arr_use)

def finite_min(arr, *, if_empty=np.nan):
    '''returns min of all the finite values of arr. (return if_empty if arr has no finite values.)'''
    return finite_op(arr, np.min, if_empty=if_empty)

def finite_mean(arr, *, if_empty=np.nan):
    '''returns mean of all the finite values of arr.  (return if_empty if arr has no finite values.)'''
    return finite_op(arr, np.mean, if_empty=if_empty)

def finite_max(arr, *, if_empty=np.nan):
    '''returns max of all the finite values of arr. (return if_empty if arr has no finite values.)'''
    return finite_op(arr, np.max, if_empty=if_empty)

def finite_std(arr, *, if_empty=np.nan):
    '''returns std of all the finite values of arr. (return if_empty if arr has no finite values.)'''
    return finite_op(arr, np.std, if_empty=if_empty)

def finite_median(arr, *, if_empty=np.nan):
    '''returns median of all the finite values of arr. (return if_empty if arr has no finite values.)'''
    return finite_op(arr, np.median, if_empty=if_empty)

def finite_percentile(arr, percentile, *, if_empty=np.nan):
    '''returns percentile of all the finite values of arr.  (return if_empty if arr has no finite values.)'''
    if getattr(arr, 'dtype', None) == bool:
        raise TypeError("finite_percentile does not support boolean arrays.")
    return finite_op(arr, lambda arr: np.percentile(arr, percentile), if_empty=if_empty)


''' --------------------- unique close --------------------- '''

def unique_close(arr, *, rtol=1e-5, atol=1e-8):
    '''returns unique values of arr, within relative tolerance rtol and absolute tolerance atol.
    Like np.unique(arr), but it uses np.isclose to compare values, instead of exact equality.
    '''
    result = []
    arr = np.asanyarray(arr)
    for elem in arr.flat:
        close = np.isclose(result, elem, atol=atol, rtol=rtol)
        if not np.any(close):
            result.append(elem)
    return np.array(result)


''' --------------------- indexing --------------------- '''

@format_docstring(default_rounding=DEFAULTS.FRACTIONAL_INDEX_ROUNDING)
def interprets_fractional_indexing(indexer, L=None, *, rounding=UNSET):
    '''interprets any fractional values, i.e. non-integers between -1 and 1.
    returns indexer in "canonical" form, no longer containing any fractional values.
    If indexer contains no fractional values, return indexer unchanged.

    indexer: None, int, slice, iterable, or non-integer between -1 and 1.
        indexer to interpret; might contain fractional value(s), i.e. non-integers between -1 and 1.
        any fractional values will be converted to value*(L+1), then rounded to an integer.
            rounding method is determined by `rounding`, unless indexer is a slice,
            in which case rounding for start & stop is handled differently.
        fractional values might appear in any of the following ways:
            - as an individual float, e.g. 0.3
            - as a float inside an iterable, e.g. [0, 0.25, 0.5, 0.75, 1]
            - as a float inside a slice, e.g. slice(0, -0.1, 0.02)
        "fractional" is tested via (-1 < value < 1) and (value != 0).
    L: None or int
        length of the object being indexed. Required if any fractional values provided.
        None --> if any fractional values provided, raise InputMissingError.
        L must be >= 2. [TODO] handle L=1 case (have all fractional values imply 0).
    rounding: UNSET, 'round', 'floor', 'ceil' or 'int'
        method to use for rounding fractional values to integers, except for slice.start or slice.stop.
        UNSET --> use DEFAULTS.FRACTIONAL_INDEX_ROUNDING (default: {default_rounding})
        'round' --> as per builtins.round(). round to nearest integer, ties toward even integers.
        'int' --> as per builtins.int(). round towards 0.
        'floor' --> as per math.floor(). round towards negative infinity.
        'ceil' --> as per math.ceil(). round towards positive infinity.
    
    Rounding for slice.start and start.stop will ALWAYS include all indices between start*(L-1) and stop*(L-1)
        For positive (or None) step, this means: round all numbers toward +infinity.
            Examples (floats here are after multiplying input by L-1, e.g. 0.1 * 153 --> 15.3):
                slice(15.3, 29.2, 1)    --> slice(16, 30, 1)     # 15.3, [16, 17, ..., 29], 29.2
                slice(-29.2, -15.3, 1)  --> slice(-29, -15, 1)   # -29.2, [-29, -28, ..., -16], -15.3
                slice(15.3, -15.3, 1)   --> slice(16, -15, 1)    # 15.3, [16, 18, ..., -17, -16], -15.3
                slice(-29.2, 100.1, 1)  --> slice(-29, 101, 1)   # -29.2, [-29, -28, ..., 99, 100], 100.1
        For negative step, this means: round all numbers toward -infinity.
            Examples (floats here are after multiplying input by L-1, e.g. 0.1 * 153 --> 15.3):
                slice(29.2, 15.3, -1)   --> slice(29, 15, -1)    # 29.2, [29, 28, ..., 17, 16], 15.3
                slice(-15.3, -29.2, -1) --> slice(-16, -30, -1)  # -15.3, [-16, -17, ..., -28, -29], -29.2
                slice(-15.3, 15.3, -1)  --> slice(-16, 15, -1)   # -15.3, [-16, -17, ..., 16, 15], 15.3
                slice(100.1, -29.2, -1) --> slice(100, -30, -1)  # 100.1, [100, 99, ..., -28, -29], -29.2
    Rounding mode for slice.step is determined by `rounding`, applying to step*L (not L-1).
        however if this causes step=0, instead use step=1 or step=-1, based on sign of original step.

    --- Examples ---
        # [TODO] update these when considering N = L-1 is what is actually being multiplied.
        interprets_fractional_indexing(0.5, L=10)  --> 5
        interprets_fractional_indexing(slice(0.2, 0.9, 0.1), L=10)  --> slice(2, 9, 1)
        interprets_fractional_indexing([0.2, 0.5, -3, 8, 7, 0.9], L=10)  --> [2, 5, -3, 8, 7, 9]
        interprets_fractional_indexing(0.23, L=10, rounding='int')  --> 2
        interprets_fractional_indexing(0.23, L=10, rounding='ceil')  --> 3
    '''
    if indexer is None:
        return indexer
    if np.ndim(indexer) > 1:
        raise NotImplementedError("interprets_fractional_indexing does not yet support multi-dimensional indexers.")
    def is_fractional(value):
        return (-1 < value < 1) and (value != 0)
    def check_L_because_it_is_required():
        if L is None:
            raise InputMissingError('L is required if any fractional values are provided.')
        if L <= 1:
            raise ValueError(f"fractional indexing not supported when L <= 1; got L={L}")
    if rounding is UNSET:
        rounding = DEFAULTS.FRACTIONAL_INDEX_ROUNDING
    if L is not None:
        N = L - 1
    if isinstance(indexer, slice):
        start, stop, step = indexer.start, indexer.stop, indexer.step
        step_is_frac = ((step is not None) and is_fractional(step))
        start_is_frac = ((start is not None) and is_fractional(start))
        stop_is_frac = ((stop is not None) and is_fractional(stop))
        if not start_is_frac and not stop_is_frac and not step_is_frac:
            return indexer  # no fractional values; return indexer unchanged.
        #else:
        check_L_because_it_is_required()
        if start_is_frac or stop_is_frac:
            if (step is None) or (step > 0):
                if start_is_frac: start = round_to_int(start * N, mode='ceil')
                if stop_is_frac: stop = round_to_int(stop * N, mode='ceil')
            else:
                if start_is_frac: start = round_to_int(start * N, mode='floor')
                if stop_is_frac: stop = round_to_int(stop * N, mode='floor')
        if step_is_frac:
            orig_step = step
            step = round_to_int(step * L, mode=rounding)
            if step == 0:
                step = 1 if (orig_step > 0) else -1
        return type(indexer)(start, stop, step)
    # else
    try:
        iter(indexer)
    except TypeError:
        is_iterable = False
    else:
        is_iterable = True
    if is_iterable:
        indexer_type = type(indexer)
        result = [val for val in indexer]
        fractional = [is_fractional(val) for val in result]
        if any(fractional):
            check_L_because_it_is_required()
            result = [(round_to_int(val * N, mode=rounding) if f else val) for val, f in zip(result, fractional)]
            return indexer_type(result)
        else:
            return indexer  # no fractional values; return indexer unchanged.
    else:  # not iterable
        if is_fractional(indexer):
            check_L_because_it_is_required()
            return round_to_int(indexer * N, mode=rounding)
        else:
            return indexer
    assert False, "unreachable code"

def ndindex_array(shape):
    '''returns an array of indexes, where each element is its own np.ndindex.
    For 0D, crash with DimensionalityError.
    For 1D, result[i] == i
        (Do not use tuple results for 1D. Also, optimize dtype to np.min_scalar_type.)
        E.g. ndindex_array((3,)) --> np.array([0,1,2], dtype=uint8).
    For 2D+, result[index] == index, where index is a tuple.
        E.g. in 2D, result[i,j] == (i,j). And, ndindex_array((2,3)) -->
            np.array([[(0,0),(0,1),(0,2)], [(1,0),(1,1),(1,2)]], dtype=object)
    '''
    if not is_iterable(shape):
        shape = [shape]
    if len(shape) == 0:
        raise DimensionalityError('ndindex_array does not support 0D arrays.')
    elif len(shape) == 1:
        result = np.arange(shape[0], dtype=np.min_scalar_type(shape[0]))
    else:  # 2D+
        result = np.empty(shape, dtype=object)
        for index in np.ndindex(shape):
            result[index] = index
    return result


''' --------------------- array from nested list --------------------- '''

def looks_flat(x):
    '''returns whether x looks flat, i.e. looks like an iterable with no internal layers.
    Only checks x[0]. (If x is not iterable, raise TypeError. If len(x)==0, return True).
    Note: might behave unexpectedly for strings.
    '''
    if not is_iterable(x):
        raise TypeError(f'looks_flat(x) expected iterable x but got x={x}')
    if len(x)==0:
        return True
    return not is_iterable(x[0])

def nest_shape(nested_list, is_element=looks_flat):
    '''returns the implied shape for a numpy object array constructed from nested_list.
    Only considers 0th element of list(s) (recursively, as needed). Stops when reaching an is_element(obj).
    To avoid infinite recursion, raise ValueError if current[0] is current (e.g. with strings).
    (To properly handle strings, provide a more sophisticated is_element than the default value.)
    
    is_element: callable of one input
        if is_element(obj), stop going deeper into lists.
        otherwise, append len(obj) to result, then inspect obj[0].
        (to start, use obj = nested_list.)
        default (looks_flat) stops when obj is iterable but obj[0] is not.
    '''
    shape = []
    current = nested_list
    while not is_element(current):
        if current[0] is current:
            raise DimensionalityError('nest_shape failed: obj[0] is obj. Crash to avoid infinite loop.')
        l = len(current)
        shape.append(l)
        current = current[0]
    return shape


''' --------------------- wrap 1D list --------------------- '''

def wraplist(x, wraprow=None, wrapcol=None, *, fill=None, dtype=object):
    '''wrap 1D list into a 2D array. Fill missing elements with fill.

    Must provide wraprow or wrapcol, but not both:
    wraprow: None or int
        result rows will have this length. result goes left to right then top to bottom.
        E.g. [1,2,3,4,5,6,7,8,9], wraprow=4 --> [[1,2,3,4], [5,6,7,8], [9,fill,fill,fill]]
    wrapcol: None or int
        result cols will have this length. result goes top to bottom then left to right.
        E.g. [1,2,3,4,5,6,7,8,9], wrapcol=4 --> [[1,5,9], [2,6,fill], [3,7,fill], [4,8,fill]]
    fill: None or any value
        fill missing elements with this value.
    dtype: any valid specifier for a numpy dtype
        dtype for resulting numpy array.
    '''
    if wraprow is None and wrapcol is None:
        raise InputMissingError('wraplist requires either wraprow or wrapcol to be provided.')
    if wraprow is not None and wrapcol is not None:
        raise InputConflictError('wraplist requires either wraprow or wrapcol, not both.')
    Lx = len(x)
    # var names below as if we are using wraprow. Code works for wrapcol too though.
    Lrow = wraprow if (wraprow is not None) else wrapcol
    Lcol = Lx // Lrow + (0 if (Lx % Lrow)==0 else 1)
    # shape=(Lcol, Lrow) because Lrow = number of cols. E.g. Lrow=7 means there are 7 cols.
    result = np.full((Lcol, Lrow), fill, dtype=dtype)
    result.flat[:Lx] = x   # without [:Lx], this wraps around unexpectedly...
    if wrapcol is not None:
        result = result.T
    return result

def ndenumerate_nonNone(arr, none=None):
    '''np.ndenumerate(arr) but skip wherever value is none (compared via 'is', not '==').'''
    for index, value in np.ndenumerate(arr):
        if value is not none:
            yield index, value




''' --------------------- types --------------------- '''

def np_dtype_object_to_str(array):
    '''convert a numpy array of dtype==object into array of str(x) for all x in array.'''
    try:
        return array.astype(str)
    except ValueError:  # e.g. ValueError: setting an array element with a sequence
        pass  # handled below -- call str() on each element individually
    result = np.empty(array.shape, dtype=object)  # can't dtype=str here because it would do '<U1'.
    for index, value in np.ndenumerate(array):
        result[index] = str(value)
    return result.astype(str)



"""
File Purpose: equation of state (EOS) and radiation tables for single-fluid MHD analysis;
Function of eperm (internal energy per unit mass) and r (mass density) --> quantity,
    for various quantities like pressure, temperature, entropy, opacity.

Optionally, can provide e (internal energy density) instead of eperm;
    in that case will convert internally via eperm = e / r.
"""

import os

import numpy as np
import xarray as xr

from .elements import ElementList
from ..errors import (
    DimensionalityError,
    FileContentsError,
    InputConflictError, InputMissingError,
)
from ..tools import (
    alias, simple_property,
    format_docstring,
    UNSET, NO_VALUE,
    using_attrs,
    read_idl_params_file, attempt_literal_eval,
    DictlikeFromKeysAndGetitem,
    product, is_iterable,
    xarray_is_sorted, xarray_interp_inverse,
)
from ..units import UnitsHaver


''' --------------------- eperm rho tables --------------------- '''

_paramdocs_ertable = {
    # making erTable
    'xtable': '''UNSET or xarray.DataArray, probably 2D
        must have dims & coords for 'eperm' and 'r'.
        eperm = internal energy per unit mass; r = mass density.
        UNSET --> will call self.load_xtable() to load it.''',
    'quantity': '''None or str
        name of quantity, if known.''',
    'ln_quant': '''bool
        whether this quantity's values are in ln-space. E.g. ln(P) instead of P.''',
    'array': '''array, at least 2D
        first two dims must correspond to r (dim0), and eperm (dim1)
        additional dims will be unlabeled, E.g. dim_2, dim_3, dim_4, ...''',
    'kw_er_coords': '''epermcoords, rcoords: 1D array
        coords for eperm (energy per unit mass) and r (mass density).
        Units should be 'raw' units.''',
    # erTable from memmap
    'extra_dims': '''extra_dims: None, list of str, or dict
        if provided, will be used as names for additional xarray dims.
        if dict, will be used as names and coords for additional dims.''',
    'data_shape': '''None or tuple
        shape of a single array in the memmap file.
        None --> infer shape = (len(epermcoords), len(rcoords))''',
    'data_dtype': '''data_dtype: np.dtype or something which implies np.dtype
        dtype of the data in memmap file''',
    'data_order': ''''C' or 'F'
        order of the data in memmap file. 'F' is Fortran order.''',
    'scale_quant': '''number, default 1.
        scale quantity by this value immediately after reading from memmap.
        (always corresponds to the not-ln'd quant, even if ln_quant=True).''',
    'scale_quants': r'''dict
        {quantity: scaling factor} for each quantity in file. Default scaling factor is 1.
        quantity value * scaling factor should always convert to 'raw' units.
        (always corresponds to the not-ln'd quant, even if ln_quant=True).''',
    # calling interp
    'r': '''number or array
        evaluate table at these value(s) of r (mass density).''',
    'eperm': '''UNSET, number, or array
        evaluate table at these value(s) of eperm (internal energy per unit mass).
        must provide eperm or e, but not both.''',
    'e': '''UNSET, number, or array
        evaluate table at these value(s) of e (internal energy density).
        since table uses eperm and r, internally convert e to eperm via eperm = e / r.
        must provide eperm or e, but not both.''',
    'ln': '''bool
        whether to report result in ln-space, e.g. ln(P) instead of P.''',
    'keep_er_coords': '''bool
        whether to keep eperm and r as coords in result.''',
    # handling units
    'u': '''None or UnitsManager
        used to calculate scale_quants to convert to 'raw' units.
        None --> all scale quants will be 1.''',
    'units_input': '''None, 'raw', 'cgs', or 'si'
        inputs unit system (for eperm, r, and quantity).
        ignored if u is trivial, else required.''',
}

@format_docstring(**_paramdocs_ertable)
class erTable():
    '''lookup table from eperm and r to quantity. Units are always 'raw' units.
    Call self(r=r, eperm=eperm) to get value of quantity at eperm and r.
        (equivalent: self.interp(r=r, eperm=eperm). See help(self.interp) for more details.)

    xtable: {xtable}
    quantity: {quantity}
    ln_quant: {ln_quant}
    '''
    def __init__(self, xtable=UNSET, quantity=None, *, ln_quant=False, **kw_super):
        if xtable is not UNSET:
            xtable = self.labeled_xtable(xtable, quantity=quantity)
        self.xtable = xtable
        self.quantity = quantity
        self.ln_quant = ln_quant

    @staticmethod
    def labeled_xtable(xtable, *, quantity=None, units='raw'):
        '''return copy of xarray.DataArray, labeled with name=quantity and attrs['units']='raw'
        Does not make copy if xtable is already labeled appropriately.
        '''
        if (xtable.name == quantity) and ('units' in xtable.attrs) and (xtable.attrs['units'] == units):
            return xtable
        result = xtable.copy()
        result.name = quantity
        result.attrs['units'] = units
        return result

    # # # SUBCLASS COULD IMPLEMENT # # #
    def load_xtable(self):
        '''load xtable. Here, raises NotImplementedError.
        Subclasses which do not implement this should provide xtable during initialization.
        '''
        raise NotImplementedError(f'implement {type(self).__name__}.load_xtable() or provide xtable during init.')

    # # # OTHER WAYS TO CREATE INSTANCES OF THIS CLASS # # #
    @staticmethod
    @format_docstring(**_paramdocs_ertable, sub_ntab=1)
    def xtable_from_array(array, *, epermcoords, rcoords, quantity=None, extra_dims=None):
        '''create xtable from unlabeled array and coords.
        no unit conversions are performed here; provide inputs in the desired unit system.

        array: {array}
        {kw_er_coords}
        quantity: {quantity}
        extra_dims: {extra_dims}
        '''
        dims = ['eperm', 'r']
        if extra_dims is not None:
            dims.extend(list(extra_dims))
        for i in range(len(dims), np.ndim(array)):  # if missing any dims use unnamed.
            dims.append(f'dim_{i}')
        coords = {'r': rcoords, 'eperm': epermcoords}
        if isinstance(extra_dims, dict):
            coords.update(extra_dims)
        return xr.DataArray(array, dims=dims, coords=coords, name=quantity)

    @classmethod
    @format_docstring(**_paramdocs_ertable, sub_ntab=1)
    def from_array(cls, array, *, epermcoords, rcoords,
                   quantity=None, extra_dims=None, ln_quant=False, **kw_init):
        '''create reTable from unlabeled array, and coords.
        
        array: {array}
        {kw_er_coords}
        quantity: {quantity}
        ln_quant: {ln_quant}
        extra_dims: {extra_dims}
        '''
        kw = dict(epermcoords=epermcoords, rcoords=rcoords,
                  quantity=quantity, extra_dims=extra_dims)
        kw_cls = dict(ln_quant=ln_quant, **kw_init)
        xtable = cls.xtable_from_array(array, **kw)
        return cls(xtable, **kw_cls)

    # # # XTABLE UNITS CHECKING # # #
    @property
    def xtable(self):
        '''xarray.DataArray, probably 2D; must have dims for eperm and r.
        if self.xtable was UNSET, self.load_xtable() first. Units are always 'raw' units.
        '''
        result = self._xtable
        if result is UNSET:
            result = self.load_xtable()
        self._xtable = result
        return result
    @xtable.setter
    def xtable(self, value):
        self._xtable = value

    # # # EVALUATE # # #
    __call__ = alias('interp')

    @format_docstring(**_paramdocs_ertable, sub_ntab=1)
    def interp(self, *, r, eperm=UNSET, e=UNSET, ln=False, keep_er_coords=False, **kw_xarray_interp):
        '''return values of quantity at eperm and r.

        r: {r}
        eperm: {eperm}
        e: {e}
        ln: {ln}
        keep_er_coords: {keep_er_coords}
        '''
        if eperm is UNSET:
            if e is UNSET:
                raise InputMissingError('provide eperm or e.')
            else:
                eperm = e / r
        elif (e is not UNSET):  # and eperm is not UNSET
            raise InputConflictError('cannot provide both eperm and e; must provide only one of them.')
        kw_xarray_interp.setdefault('assume_sorted', self.coords_are_sorted)  # [EFF] faster if sorted!
        result = self.xtable.interp(eperm=eperm, r=r, **kw_xarray_interp)  # xarray.DataArray.interp(...)
        if not keep_er_coords:
            result = result.drop_vars(('eperm', 'r'))
        # exponentiate (or take ln) if necessary
        if self.ln_quant and (not ln):
            result = np.exp(result)
        elif (not self.ln_quant) and ln:
            result = np.log(result)
        return result

    coords_are_sorted = simple_property('_coords_are_sorted', setdefaultvia='can_assume_sorted',
            doc='''whether eperm and r coords (of self.xtable) are sorted (in non-decreasing order).
            Cache result for efficiency, assumes coords are not changed later.''')

    def can_assume_sorted(self):
        '''return whether eperm and r coords are sorted (in non-decreasing order).
        If True, can assume sorted for xarray interp calls, which improves efficiency signficantly.
        '''
        eperm = self.xtable.coords['eperm']
        r = self.xtable.coords['r']
        return xarray_is_sorted(eperm) and xarray_is_sorted(r)

    def interp_inverse(self, quantity, *, r=UNSET, eperm=UNSET, ln=False, **kw_interp_inverse):
        '''return values of r or eperm, given quantity values and other coord's values.

        quantity: number or array
            values of quantity.
        r, eperm: UNSET, number, or array
            values of r, eperm. (r=mass density; eperm=internal energy per unit mass==e/r)
            must provide exactly one of these.
        ln: bool
            whether the values of quantity input to this function are in ln-space.

        any additional kwargs are passed to xarray_interp_inverse.
        '''
        # bookkeeping
        if r is UNSET and eperm is UNSET:
            raise InputMissingError('provide r or eperm.')
        elif r is not UNSET and eperm is not UNSET:
            raise InputConflictError('provide exactly one of r or eperm.')
        interpto = {'r': r} if (r is not UNSET) else {'eperm': eperm}
        _output = 'r' if r is UNSET else 'eperm'
        kw_interp_inverse.setdefault('assume_sorted', self.coords_are_sorted)  # [EFF] faster if sorted!
        # exponentiate (or take ln) if necessary
        if self.ln_quant and (not ln):
            quantity = np.log(quantity)
        elif (not self.ln_quant) and ln:
            quantity = np.exp(quantity)
        # bookkeeping for quantity name stuff:
        xtable = self.xtable
        if xtable.name is None:
            xtable = xtable.rename('_unnamed_xtable_quantity')
        interpto[xtable.name] = quantity
        # get result
        return xarray_interp_inverse(self.xtable, interpto, **kw_interp_inverse)

    # # # DISPLAY # # #
    def __repr__(self):
        contents = [f'shape={self.xtable.shape}']
        if self.quantity is not None:
            contents.append(f'quantity={self.quantity!r}')
        if self.ln_quant:
            contents.append(f'ln_quant={self.ln_quant!r}')
        return f'{type(self).__name__}({", ".join(contents)})'


@format_docstring(**_paramdocs_ertable)
class erTableFromMemmap(erTable):
    '''lookup table from eperm and r to quantity. Table loaded from a memmap file.
    Internally, all values should be in 'raw' units (see scale_quant kwarg for more details).

    Call self(r=r, eperm=eperm) to get value of quantity at eperm and r.
        (equivalent: self.interp(r=r, eperm=eperm). See help(self.interp) for more details.)

    filename: str
        name of file. Internally stored as abspath.
    quantity: {quantity}
    {kw_er_coords}
    index_in_file: int
        n such that this table is the nth table in the memmap file.
    ln_quant: {ln_quant}
    scale_quant: {scale_quant}
    extra_dims: {extra_dims}
        use this when memmap arrays have more dims than just eperm and r.
    data_shape: {data_shape}
        if extra_dims provided with coords, include those in inferred shape too.
    data_dtype: {data_dtype}
    data_order: {data_order}
    
    [EFF] loading is "lazy"; wait to load until first time the table is actually accessed.
    '''
    def __init__(self, filename, quantity=None, *,
                 epermcoords, rcoords, index_in_file,
                 ln_quant=False, scale_quant=1,
                 extra_dims=None, data_shape=None,
                 data_dtype=np.float32, data_order='F',
                 **kw_super):
        self.filename = os.path.abspath(filename)
        self.rcoords = rcoords
        self.epermcoords = epermcoords
        self.index_in_file = index_in_file
        self.scale_quant = scale_quant
        self.extra_dims = extra_dims
        self.data_shape = data_shape
        self.init_infer_data_shape()
        self.data_dtype = np.dtype(data_dtype)
        self.data_order = data_order
        super().__init__(xtable=UNSET, # xtable not loaded yet!
                         quantity=quantity, ln_quant=ln_quant, **kw_super)

    def init_infer_data_shape(self):
        '''infer data_shape from self.data_shape and self.extra_dims;
        set self.data_shape to the inferred value.'''
        data_shape = self.data_shape
        extra_dims = self.extra_dims
        if data_shape is None:
            data_shape = [len(self.epermcoords), len(self.rcoords)]
            if extra_dims is not None:
                if isinstance(extra_dims, dict):
                    data_shape.extend([len(coords) for coords in extra_dims.values()])
                else:
                    errmsg = f'cannot infer data_shape when extra_dims provided without coord info.'
                    raise InputConflictError(errmsg)
            data_shape = tuple(data_shape)
        elif (extra_dims is not None) and (len(data_shape) < 2 + len(extra_dims)):
            errmsg = f'data_shape dimensionality {data_shape} inconsistent with number of extra_dims.'
            raise DimensionalityError(errmsg)
        self.data_shape = data_shape
        return data_shape

    def _load_table_array(self):
        '''load table array from memmap file.'''
        array_size_nbytes = product(self.data_shape) * self.data_dtype.itemsize
        offset_nbytes = array_size_nbytes * self.index_in_file
        result = np.memmap(self.filename, offset=offset_nbytes,
                           mode='r',  # read-only; never alters existing files!
                           shape=self.data_shape,
                           dtype=self.data_dtype,
                           order=self.data_order,
                           )
        result = np.asarray(result)  # convert to numpy array.
        return result

    def load_xtable(self):
        '''load table xarray from memmap file. scale quant by self.scale_quant if not 1.
        save result to self.xtable, also return it.
        '''
        array = self._load_table_array()
        kw = dict(epermcoords=self.epermcoords, rcoords=self.rcoords,
                  quantity=self.quantity, extra_dims=self.extra_dims)
        xtable = self.xtable_from_array(array, **kw)
        if self.scale_quant != 1:
            if self.ln_quant:
                # result = log(exp(xtable) * factor)
                #       == log(exp(xtable)) + log(factor)
                #       == xtable + log(factor)
                xtable = xtable + np.log(self.scale_quant)
            else:
                xtable = xtable * self.scale_quant
        xtable = self.labeled_xtable(xtable, quantity=self.quantity)
        self.xtable = xtable
        return xtable


class erTableManager(dict):
    '''manages er tables. Dict of {quantity (str): erTable}'''
    er_table_from_memmap_cls = erTableFromMemmap

    __call__ = alias('interp')

    @format_docstring(**_paramdocs_ertable, sub_ntab=1)
    def interp(self, var, *, r, eperm=UNSET, e=UNSET, ln=False,
               keep_er_coords=False, **kw_xarray_interp):
        '''return values of var at eperm and r.

        var: str
            name of quantity to interpolate.
        r: {r}
        eperm: {eperm}
        e: {e}
        ln: {ln}
        keep_er_coords: {keep_er_coords}
        '''
        table = self[var]
        kw = dict(ln=ln, keep_er_coords=keep_er_coords, **kw_xarray_interp)
        return table.interp(r=r, eperm=eperm, e=e, **kw)

    # # # FROM FILE # # #
    @classmethod
    @format_docstring(**_paramdocs_ertable, sub_ntab=1)
    def from_file(cls, memmap_file, *, epermcoords, rcoords,
                  quantities, ln_quants=[], scale_quants={}, extra_dims=None,
                  data_shape=None, data_dtype=np.float32, data_order='F',
                  **kw_init_er_table):
        '''create erTableManager from numpy memmap file with multiple tables.
        
        memmap_file: str
            path to memmap file.
        {kw_er_coords}
        quantities: list of str
            names of quantities in file.
            each quantitiy in file should be an ND array with dims eperm, r, and any extra_dims.
        ln_quants: list of str
            names of quantities whose values in file are in ln-space.
        scale_quants: {scale_quants}
        extra_dims: {extra_dims}
        data_shape: {data_shape}
        data_dtype: {data_dtype}
        data_order: {data_order}
        '''
        kw = dict(epermcoords=epermcoords, rcoords=rcoords,
                  extra_dims=extra_dims, data_shape=data_shape,
                  data_dtype=data_dtype, data_order=data_order,
                  **kw_init_er_table)
        file = os.path.abspath(memmap_file)
        manager = cls()
        for i, var in enumerate(quantities):
            ln = var in ln_quants
            table = cls.er_table_from_memmap_cls(file, var, index_in_file=i, ln_quant=ln,
                                                 scale_quant=scale_quants.get(var, 1),
                                                 **kw)
            manager[var] = table
        return manager

    # # # DISPLAY # # #
    def __repr__(self):
        return f'{type(self).__name__}(keys={list(self.keys())})'


@format_docstring(**_paramdocs_ertable)
def eos_file_tables(eos_file, *, epermcoords, rcoords, u=None):
    '''return erTableManager dict (of {{var: Equation Of State lookup table}}), for vars in file.

    file should have quantities:
        'P' (pressure),
        'T' (temperature),
        'ne' (electron number density),
        'kappaR' (Rosseland opacity).
    file values of P, ne, and kappaR should be in ln-space (e.g. ln(P) instead of P).
    file should have values in cgs
    (despite note in helita implying ne would be in si... I'm 99% sure it is in cgs here.
        e.g. ne close to n (== r / (amu * self.M_AMU_FLUID)) in the corona,
        where plasma should be mostly ionized.
        As opposed to ne 1e-6 times smaller than n, if ne was truly in si in table.)

    eos_file: str
        path to eos table file.
    {kw_er_coords}
    u: {u}
    '''
    kw_eos = dict(
        # properties of the "standard eos table" file.
        #  to load eos tables from file with nonstandard properties,
        #  write a different function but feel free to use this code as an example.
        quantities = ['P', 'T', 'ne', 'kappaR'],
        ln_quants = ['P', 'ne', 'kappaR'],
        data_type = np.float32,
        data_order = 'F',
        )
    if u is None:
        scaling = dict()
    else:
        scaling = {
            'P': u('pressure', 'raw', convert_from='cgs'),
            'T': u('temperature', 'raw', 'cgs'),
            'ne': u('number_density', 'raw', 'cgs'),
            'kappaR': u('opacity', 'raw', 'cgs'),
            }
    kw_eos['scale_quants'] = scaling
    kw = dict(epermcoords=epermcoords, rcoords=rcoords)
    return erTableManager.from_file(eos_file, **kw, **kw_eos)


@format_docstring(**_paramdocs_ertable)
def rad_file_tables(rad_file, *, epermcoords, rcoords, radbins, u=None):
    '''return erTableManager dict (of {{var: Radiation lookup table}}), for vars in file.

    file should have quantities:
        'pscatter' (probability of scattering)
        'emtherm' (thermal emission)
        'opacity' (opacity)
    file values should all (?) be in ln-space (e.g. ln(emtherm) instead of emtherm). [TODO] check
    file should have values in cgs. [TODO] check

    rad_file: str
        path to rad table file.
    {kw_er_coords}
    radbins: int or 1D array-like
        number of radiation bins or coordinates of radiation bins.
    u: {u}
    '''
    if not is_iterable(radbins):
        radbins = np.arange(radbins)
    extra_dims = {'radbin': radbins}
    kw_rad = dict(
        # properties of the "standard rad table" file.
        #  to load rad tables from file with nonstandard properties,
        #  write a different function but feel free to use this code as an example.
        quantities = ['pscatter', 'emtherm', 'opacity'],
        ln_quants = ['pscatter', 'emtherm', 'opacity'],  # [TODO] check
        extra_dims = extra_dims,
        data_type = np.float32,
        data_order = 'F',
        )
    if u is None:
        scaling = dict()
    else:
        scaling = {
            'opacity': u('opacity', 'raw', convert_from='cgs'),
            # [TODO] units for pscatter and emtherm? 
            }
    kw = dict(epermcoords=epermcoords, rcoords=rcoords)
    return erTableManager.from_file(rad_file, **kw, **kw_rad)


''' --------------------- directly related to tabparam.in --------------------- '''

@format_docstring(**_paramdocs_ertable)
class erTabInputManager(DictlikeFromKeysAndGetitem, UnitsHaver):
    '''manages tables, based on tab input file.
    All tables and table coords here are in 'raw' units.
    self is dict-like, with keys quantity names, values erTable objects.
        (alternatively, use self.tables for dict like self.)
    self.elements is an ElementList based on the elements info in tab input file.

    filename: str
        path to tab input file, e.g. tab_param.in. Internally stored as abspath.
        This should be an idl-like file which includes params for tables:
            EOSTableFile: str, path to eos table file
            RhoEiRadTableFile: str, path to radiation table file
            nRhoBin: int, number of bins for r (mass density)
            RhoMin, RhoMax: numbers, r [cgs units] for smallest and largest bins
            nEiBin: int, number of bins for eperm (internal energy per unit mass)
            EiMin, EiMax: numbers, eperm [cgs units] for smallest and largest bins
            nRadBins: int, number of radiation bins in the radiation tables.
        It should also include params for elements (each as a string separated by spaces):
            cel: elements names. Internally stored in title case (e.g. 'H', 'He').
            abund: abundances of elements: A(elem) = 12 + log10(n(elem) / n(H)).
            aweight: masses of elements [amu].
    u: {u}
        example: if this TabInputManager is associated with a BifrostCalculator,
            probably will have u = the BifrostUnitsManager from that BifrostCalculator.
    ionize_ev: str or dict
        first ionization potentials of elements [eV]. E.g., {{'H': 13.6}}.
        affects self.elements; irrelevant to tables.
        str --> use ElementList.DEFAULTS['ionize_ev'][ionize_ev]
    '''
    er_table_manager_cls = erTableManager
    element_list_cls = ElementList

    def __init__(self, filename, *, u=None, ionize_ev='physical', **kw_super):
        self.filename = os.path.abspath(filename)
        self.params = read_idl_params_file(self.filename, eval=True)
        self.ionize_ev = ionize_ev
        # units
        super().__init__(u=u, **kw_super) 
        self.init_tables()  # makes self.tables
        self.make_element_list()  # makes self.elements

    def init_tables(self):
        '''initialize self.tables and return result'''
        kw_coords = self.get_coords(for_kw=True)
        # eos tables
        eosfile = self.eosfile
        eos = eos_file_tables(eosfile, **kw_coords, u=self.u)
        # rad tables
        radfile = self.radfile
        radbins = self.n_radbins
        rad = rad_file_tables(radfile, **kw_coords, u=self.u, radbins=radbins)
        # result
        result = eos
        result.update(rad)
        self.tables = result
        return result

    def make_element_list(self, *, ionize_ev=None, set=True):
        '''set & return self.elements = ElementList based on params in self.
        ionize_ev: None, str, or dict
            first ionization potentials of elements [eV]. E.g., {'H': 13.6}.
            str --> use ElementList.DEFAULTS['ionize_ev'][ionize_ev]
            None --> use self.tabin.ionize_ev (default: 'physical')
        set: bool
            whether to set self.elements to the result. Default True.
        '''
        if ionize_ev is None:
            ionize_ev = self.ionize_ev
        cel = self.params['cel'].split()
        abund = self.params['abund'].split()
        aweight = self.params['aweight'].split()
        # check lengths
        if not (len(cel) == len(abund) == len(aweight)):
            raise FileContentsError('cel, abund, and aweight imply different numbers of elements')
        names = [e.title() for e in cel]
        m = {e: attempt_literal_eval(v) for e, v in zip(names, aweight)}
        abundance = {e: attempt_literal_eval(v) for e, v in zip(names, abund)}
        result = self.element_list_cls.from_names(names, m=m, abundance=abundance, ionize_ev=ionize_ev)
        if set: self.elements = result
        return result

    # # # RELEVANT PATHS # # #
    @property
    def dirname(self):
        '''directory containing tabinputfile. Equivalent: os.path.dirname(self.filename)'''
        return os.path.dirname(self.filename)

    @property
    def eosfile(self):
        '''filepath to eos table file.'''
        return os.path.join(self.dirname, self.params['EOSTableFile'])

    @property
    def radfile(self):
        '''filepath to radiation table file.'''
        return os.path.join(self.dirname, self.params['RhoEiRadTableFile'])

    # # # RELEVANT QUANTITIES # # #
    def get_r_coords(self):
        '''return r coords in 'raw' units. Assumes params from paramfile are in cgs units.'''
        rmin = self.params['RhoMin'] * self.u('r', 'raw', convert_from='cgs')
        rmax = self.params['RhoMax'] * self.u('r', 'raw', convert_from='cgs')
        Nr = self.params['nRhoBin']
        return np.logspace(np.log10(rmin), np.log10(rmax), Nr)

    def get_eperm_coords(self):
        '''return eperm coords in 'raw' units. Assumes params from paramfile are in cgs units.'''
        emin = self.params['EiMin'] * self.u('energy_density r-1', 'raw' ,convert_from='cgs')
        emax = self.params['EiMax'] * self.u('energy_density r-1', 'raw', convert_from='cgs')
        Ne = self.params['nEiBin']
        return np.logspace(np.log10(emin), np.log10(emax), Ne)

    def get_coords(self, *, for_kw=False):
        '''return dict of eperm coords and r coords, in 'raw' units.
        Assumes params from paramfile are in cgs units.
        for_kw: bool
            False --> result keys are 'eperm', 'r'.
            True --> result keys are 'epermcoords', 'rcoords'.
        '''
        eperm = self.get_eperm_coords()
        r = self.get_r_coords()
        if for_kw:
            return {'epermcoords': eperm, 'rcoords': r}
        else:
            return {'eperm': eperm, 'r': r}

    @property
    def n_radbins(self):
        '''return number of radbins in self.'''
        return self.params['nRadBins']

    # # # DICT-LIKE BEHAVIOR # # #
    def __getitem__(self, key):
        return self.tables[key]

    def keys(self):
        return self.tables.keys()

    # # # DISPLAY # # #
    def __repr__(self):
        return f'{self.__class__.__name__}({self.filename!r}, keys={list(self.keys())})'



"""
File Purpose: pytest configuration
"""

def pytest_collection_modifyitems(items):
    '''ordering of tests - here, puts multiprocessing tests last.'''
    module_mapping = {item: item.module.__name__ for item in items}
    default_order = 0  # default order for test modules without a specified order
    order = {
        'init_test': -100,  # run this first!
        'test_multiprocessing': 10,
        'test_eppic_tfbi': 12,
        'test_eppic_reading_misc': 15,
        'test_eppic_reading_basics': 20,
    }  # larger --> get tested later.
    items.sort(key=lambda item: order.get(module_mapping[item], default_order))



"""
File Purpose: working on movies
"""

import os
import matplotlib as mpl
import matplotlib.animation as mpl_animation
import matplotlib.pyplot as plt

from .plot_settings import PlotSettings, PlotSettingsMixin
from ..tools import (
    alias, alias_child, simple_property,
    using_attrs,
    format_docstring,
    UNSET, NO_VALUE,
    repr_simple,
    ProgressUpdater,
    Tree,
)
from ..errors import (
    PlottingNotImplementedError, PlottingAmbiguityError, PlottingNframesUnknownError,
    InputError, InputConflictError,
)

from ..defaults import DEFAULTS


''' --------------------- FuncAnimation --------------------- '''

@PlotSettings.format_docstring(doc_super=mpl_animation.FuncAnimation.__doc__)
class FuncAnimation(PlotSettingsMixin, mpl_animation.FuncAnimation):
    '''matplotlib.animation.FuncAnimation class with more methods / convenience.

    verbose: bool
        verbosity, e.g. during self.save.
    progress_callback: {progress_callback}
    fps: {fps}

    All other args and kwargs go to super(), probably mpl.animation.FuncAnimation.
    For reference, docs for mpl.animation.FuncAnimation:
    ----------------------------------------------------
    {doc_super}
    '''
    def __init__(self, fig, func, *args_super, fps=UNSET, verbose=True, **kw_super):
        if fig is None:
            raise InputError('cannot make animation when fig=None! Maybe you forgot to init_plot()?')
        super().__init__(fig, func, *args_super, fps=fps, mpl_super=['mpl.animation.FuncAnimation'], **kw_super)
        self.verbose = verbose
        # fps
        _nframes = self.plot_settings.get('frames')
        _is_small = _nframes is not UNSET and _nframes <= DEFAULTS.PLOT.NFRAMES_SMALL
        _default_fps = DEFAULTS.PLOT.FPS_SMALL if _is_small else NO_VALUE  # NO_VALUE --> use plot_settings default.
        fps = self.plot_settings.get('fps', default=_default_fps)
        if fps is not None:
            self._interval = 1000/fps   # _interval is used by some matplotlib methods.

    # # # SAVING # # #
    def get_savefile(self, filename, ext=UNSET):
        '''return abspath of filename, with ext appended if not already there.
        if ext not provided, use DEFAULTS.PLOT.MOVIE_EXT.
        '''
        if ext is UNSET: ext = DEFAULTS.PLOT.MOVIE_EXT
        if ext is not None:
            filename_has_no_ext = (os.path.splitext(filename)[1] == '')
            if filename_has_no_ext:
                if not ext.startswith('.'): ext = '.' + ext
                filename = filename + ext
        return os.path.abspath(filename)

    @PlotSettings.format_docstring(ntab=2, doc_super=mpl_animation.FuncAnimation.save.__doc__)
    def save(self, filename, writer=None, *args_super, fps=UNSET, progress_callback=UNSET, **kw_super):
        '''Save the animation as a movie file by drawing every frame.

        filename: str
            output filename, e.g. 'example_movie.mp4'.
            if no extension provided, append DEFAULTS.PLOT.MOVIE_EXT.
        writer: None, str, or MovieWriter
            writer to use. passed directly to super().save.
            if MovieWriter (i.e., not None or str), ignore any provided value for fps.
        fps: {fps}
        progress_callback: {progress_callback}

        returns abspath of the saved file.

        All other args and kwargs go to super().save, probably mpl.animation.FuncAnimation.save.
        For reference, docs for mpl.animation.FuncAnimation.save:
        ---------------------------------------------------------
        {doc_super}
        '''
        progress_callback = self.get_progress_callback(progress_callback)
        fps = self.plot_settings.get('fps', fps, default=None)
        filename_as_abspath = self.get_savefile(filename)
        if not (writer is None or isinstance(writer, str)):
            # set fps=None, to avoid RuntimeError crash (unless in DEBUG mode)
            if DEFAULTS.DEBUG >= 3:   # in debugging mode...
                pass
            else:
                fps = None
        super().save(filename_as_abspath, writer=writer, *args_super, fps=fps,
                     progress_callback=progress_callback, **kw_super)
        self._progress_updater_finalize(f'movie at {filename_as_abspath!r}')
        return filename_as_abspath

    # # # PROGRESS UPDATE / HOOK # # #
    def get_progress_callback(self, progress_callback=UNSET):
        '''return the progress_callback to use for self.save, based on input & self.verbose.
        also, first deletes self._progress_updater if it exists.
        '''
        if hasattr(self, '_progress_updater'):
            del self._progress_updater
        progress_callback = self.plot_settings.get('progress_callback', progress_callback)
        if progress_callback is UNSET:
            progress_callback = self._progress_callback_default() if self.verbose else None
        return progress_callback

    def _progress_callback_default(self):
        '''return the default progress_callback for self.save.
        also sets self._progress_updater to the new ProgressUpdater object.
        '''
        updater = ProgressUpdater(DEFAULTS.PLOT.MOVIE_PROGRESS_UPDATE_FREQ)
        self._progress_updater = updater
        return lambda i, n: updater.print(f'rendering frame {i+1} of {n}')

    def _progress_updater_finalize(self, message):
        '''finalize self._progress_updater, if it exists, with the given message.'''
        if hasattr(self, '_progress_updater'):
            self._progress_updater.finalize(message, always=self.verbose)

    # # # DISPLAY-RELATED # # #
    @format_docstring(doc_super=mpl_animation.FuncAnimation.to_jshtml.__doc__)
    def to_jshtml(self, fps=UNSET, *args_super, **kw_super):
        '''Generate HTML representation of the animation.
        fps: UNSET, None, or number
            frames per second. 
            UNSET --> default to DEFAULTS.PLOT.FPS.
                        (Or use value from self, if provided)
            None --> use matplotlib defaults.

        All other args and kwargs go to super().to_jshtml, probably mpl.animation.FuncAnimation.to_jshtml.
        For reference, docs for mpl.animation.FuncAnimation.to_jshtml:
        --------------------------------------------------------------
        {doc_super}
        '''
        fps = self.plot_settings.get('fps', fps, default=None)
        if hasattr(self, '_html_representation'):  # caching
            # delete cached representation if fps is different now.
            if (fps is not None) and (fps != getattr(self, '_fps_for_html', None)):
                del self._html_representation
        with using_attrs(self, fps=None): # temporarily set fps=None (super().save doesn't like fps when using to_jshtml)
            result = super().to_jshtml(fps=fps, *args_super, **kw_super)
        # # caching, then return result # #
        # self._html_representation = ...  # unnecessary; super().to_jshtml already caches this attr.
        self._fps_for_html = fps
        return result

    def _repr_html_(self):
        '''IPython display hook for rendering.'''
        displaying = plt.rcParams["animation.html"]
        if (displaying == 'none') and DEFAULTS.PLOT.MOVIE_REPR_INLINE_HELP:
            helpmsg = (f'Getting static _repr_html_ for {type(self)};'
                       '\nnote that you can instead view animations in-line,'
                       "\nby setting plt.rcParams['animation.html']='jshtml' (or 'html5')"
                       '\n(To disable this print statement, set DEFAULTS.PLOT.MOVIE_REPR_INLINE_HELP = False)')
            print(helpmsg)
        return super()._repr_html_()


''' --------------------- MoviePlotElement & MoviePlotNode --------------------- '''

@PlotSettings.format_docstring(ntab=1)
class MoviePlotElement(PlotSettingsMixin):
    '''updatable element in a plot. Put into an MoviePlotNode to connect update routines.

    Pneumonic:
        Element connects directly to matplotlib plotting routines;
        Node connects elements to matplotlib animation rountines.
    Element only needs to contain the minimal amount of info to plot the current data;
    Node may contain more info such as all data for the entire movie.

    Subclasses must define:
        update_data(dict) -> list of updated matplotlib Artist objects.

    INPUTS HERE:
        data: dict
            contains the data to be plotted.
            Subclasses might ask for data values as input instead of data dict,
                e.g. XarrayImagePlotElement.__init__ expects input array, not {{'array': array}}.
            But, all subclasses __init__ should define self.data = dict(...) appropriately.
            __init__ here makes a copy, to avoid unintentional changes.

        Additional kwargs can be any PlotSettings; see help(self.plot_settings) for details.
    '''
    def __init__(self, data=dict(), **kw_super):
        self.data = data.copy()
        super().__init__(**kw_super)

    # # # METHODS -- SUBCLASS SHOULD IMPLEMENT # # #
    def update_data(self, data):
        '''update the plot using relevant keys in data.
        return list of all updated matplotlib Artist objects.
        '''
        raise PlottingNotImplementedError(f'{type(self).__name__}.update_data()')


@PlotSettings.format_docstring(ntab=1)
class MoviePlotNode(PlotSettingsMixin, Tree):
    '''tree node storing a MoviePlotElement (in self.obj).

    To update repeatedly & save resulting animation, see self.save() or self.get_animator().

    To register updatable sub-elements use self.add_child;
        when updating self will also update all children.

    Subclasses must define:
        get_data_at_frame(int) -> dict

    INPUTS HERE:
        obj: UNSET, None, or MoviePlotElement
            the plot element stored in this node.
            UNSET --> didn't call init_plot.
            None --> intentionally not storing an element in this node.
        parent: None or MoviePlotNode
            if provided, the parent of this node. None -> this node has no parent.
        init_plot: {init_plot}
            ignored if obj is None; when obj is None, never call init_plot during __init__.

        Additional kwargs can be any PlotSettings; see help(self.plot_settings) for details.
    '''
    def __init__(self, obj=UNSET, *, parent=None,
                 init_plot=PlotSettings.get_default('init_plot'),
                 **kw_plot_settings):
        super().__init__(obj=obj, parent=parent, init_plot=init_plot, **kw_plot_settings)
        if self.plot_settings['init_plot'] and obj is not None:
            self.init_plot()

    # # # METHODS -- SUBCLASS SHOULD IMPLEMENT # # #
    def init_plot(self):
        '''plot for the first time; save self.obj = result, which should be a MoviePlotElement.
        Should also set self.frame = self.plot_settings['init_plot_frame']  (default 0).
        '''
        raise PlottingNotImplementedError(f'{type(self).__name__}.init_plot()')
        # example code for what should appear in the subclass:
        # self._init_plot_checks()
        # frame = self.plot_settings['init_plot_frame']
        # data = self.get_data_at_frame(frame)
        # kw_plot = ...  # probably some subset of plot_settings; depends on subclass.
        # self.obj = MoviePlotElement(data, **kw_plot)
        # self.frame = frame

    def get_data_at_frame(self, frame):
        '''return dict of data for the given frame, to be used by the MoviePlotElement at self.obj.'''
        raise PlottingNotImplementedError(f'{type(self).__name__}.get_data_at_frame()')

    def get_nframes_here(self):
        '''return the number of frames that could be in the movie, based on this node.'''
        errmsg = (f'{type(self).__name__}.get_nframes_here() not implemented, and frames not provided.\n'
                  'Provide frames=int (as kwarg or set self.frames). Or, implement get_nframes_here().')
        raise PlottingNframesUnknownError(errmsg)

    # # # PROPERTIES # # #
    fig = simple_property('_fig', setdefault=plt.gcf, doc='''the Figure on which self is / will be plotted.''')
    frame = simple_property('_frame', doc='''the currently-plotted frame''', default=None)
    plotted_data = alias_child('obj', 'data', doc='''the currently plotted data.''')

    @property
    def plotted(self):
        '''whether this node's element has actually been plotted yet.
        False before init_plot; True after. Always None if self.obj is None.
        '''
        return None if (self.obj is None) else (self.obj is not UNSET)

    # # # CHECKS # # #
    def _init_plot_checks(self):
        '''checks to run before init_plot().
        Here, checks:
            - if self.plotted, raise PlottingAmbiguityError.
            - if self.obj is None, raise PlottingAmbiguityError.
        '''
        if self.plotted:
            raise PlottingAmbiguityError('init_plot was already created! Did you mean to update instead?')
        if self.obj is None:
            raise PlottingAmbiguityError('init_plot not allowed when obj=None! (node intentionally empty)')

    # # # NUMBER OF FRAMES # # #
    def get_nframes(self):
        '''return max of get_nframes_here() for self & all descendants of self.
        If any node throws PlottingNframesUnknownError, pretend they said nframes=0.
        If all nodes throw PlottingNframesUnknownError, raise the one from self.
        '''
        nframes = 0
        try:
            nframes = self.get_nframes_here()
        except PlottingNframesUnknownError as err:
            for descendant in self.flat():
                try:
                    nframes = max(nframes, descendant.get_nframes())
                except PlottingNframesUnknownError as err_child:
                    pass
            if nframes == 0:
                raise err
        return nframes

    @property
    def frames(self):
        '''the frames that could be in the movie.
        if set to None, will use self.get_nframes() instead.
        if set to a slice, will use range(self.get_nframes())[frames] instead.
        '''
        return self._get_frames(self.plot_settings.get('frames', default=None))
    @frames.setter
    def frames(self, value):
        self.plot_settings['frames'] = value

    def _get_frames(self, value):
        '''gets frames based on value and possibly self.get_nframes() if necessary.
        if result would be None, crash instead.
        '''
        if value is None:
            value = self.get_nframes()
        if isinstance(value, slice):
            nframes = self.get_nframes()
            value = range(nframes)[value]
        return value

    # # # UPDATING # # #
    def init_plots(self, *, plotted_ok=True):
        '''init_plot for self & all descendants with non-None obj.
        plotted_ok: bool
            True --> skip node if node.plotted.
            False --> call init_plot on all nodes with non-None obj.
        '''
        if (self.obj is not None) and not (plotted_ok and self.plotted):
            self.init_plot()
        for descendant in self.flat():
            if (descendant.obj is not None) and not (plotted_ok and descendant.plotted):
                descendant.init_plot()

    def update_to_frame(self, frame):
        '''update the plot for the given frame. set self.frame=frame.
        also calls update_to_frame for all children.
        return iterable of all updated artists.
        '''
        if self.plotted == False:  # equivalent: if (not self.plotted) and (self.plotted is not None)
            raise PlottingAmbiguityError('node not yet plotted! Did you forget to call init_plot?')
        artists = []
        if (self.obj is not None) and (frame < self.get_nframes_here()):
            data = self.get_data_at_frame(frame)
            artists.extend(self.obj.update_data(data))
        for child in self.children:
            artists.extend(child.update_to_frame(frame))
        self.frame = frame
        return artists

    def __call__(self, frame):
        '''update the plot for the given frame. return iterable of all updated artists.
        Defining __call__ in this way means self is compatible for direct use by FuncAnimation.
        '''
        return self.update_to_frame(frame)

    # # # SAVING / RENDERING # # #
    @PlotSettings.format_docstring(ntab=2)
    def save(self, filename, frames=UNSET, *, fps=UNSET, blit=UNSET, **kw):
        '''save the movie to filename.
        RECOMMENDED:
            first, self.save(..., frames=N), with small N (e.g. N=5), to test movie formatting.
            Troubleshooting: if movie getting cut off,
                try plt.subplots_adjust(bottom=0.2, left=0.2, right=0.8, top=0.8),
                or even more extreme values if necessary. (0 is bottom/left edge; 1 is top/right edge.)

        frames: {frames}
        fps: {fps}
        blit: {blit}

        additional kwargs passed to FuncAnimation() or FuncAnimation.save().
        returns abspath of the saved movie.
        '''
        kw_animator = self.plot_settings.pop_mpl_kwargs('mpl.animation.FuncAnimation', kw, blit=blit, frames=frames)
        kw_animator.update(fps=self.plot_settings.get('fps', fps))  # fps goes to FuncAnimation defined here
        animator = self.get_animator(**kw_animator)
        kw_save = self.plot_settings.pop_mpl_kwargs('mpl.animation.FuncAnimation.save', kw)
        return animator.save(filename, **kw_save)

    @PlotSettings.format_docstring(ntab=2)
    def get_animator(self, *, fps=UNSET, blit=UNSET, frames=UNSET, plt_close=True, **kw_func_animation):
        '''returns FuncAnimation instance using self as func.
        Use kwarg defaults from self.plot_settings, for any kwargs not provided here.

        fps: {fps}
        blit: {blit}
        frames: {frames}
        plt_close: bool
            whether to plt.close() before returning the result.
            This is useful in Jupyter, where commonly one cell might make a plot,
                then call get_animator() to display movie in-line, but not plt.close().
                In that case, plt_close=False would display animation & plot

        [TODO] use init_func kwarg to avoid calling self twice for frame 0?
        '''
        frames = self._get_frames(self.plot_settings.get('frames', frames, default=None))
        kw = self.plot_settings.get_mpl_kwargs('mpl.animation.FuncAnimation',
                                               blit=blit, frames=frames, **kw_func_animation)
        kw.update(fps=self.plot_settings.get('fps', fps))  # fps goes to FuncAnimation defined here
        result = FuncAnimation(self.fig, func=self, **kw)
        if plt_close:
            plt.close()
        return result

    ani = alias('get_animator')   # <-- ani is defined here. (str to help codebase searches: def ani(...))

    @classmethod
    def help(cls):
        '''prints a helpful message with examples for how to use this cls'''
        msg = f'''Help for {cls}:
        Typical example use-case of an object of this type:
            obj = {cls.__name__}(...)
            obj(0)   # plot frame 0
            obj(7)   # plot frame 7
            obj.save('example_movie', frames=5)   # make an example movie; only include first 5 frames.
            obj.save('example_with_all_frames')   # make a movie with all frames.

            # or, if you want to view the movie in-line:
            plt.rcParams['animation.html'] = 'jshtml'   # or 'html5'
            ani = obj.get_animator()
            ani   # if you put this at the end of a cell in Jupyter, it will display the movie.'''
        print(msg)


''' --------------------- EmptyMovieNode & MovieOrganizerNode --------------------- '''

class EmptyMovieNode(MoviePlotNode):
    '''makes an empty MoviePlotNode; can be used to improve tree readability.
    E.g. if tree represents subplots but one has no animations on it,
        might store an EmptyMovieNode so that the indexing aligns with subplots.
    '''
    def __init__(self, **kw_super):
        super().__init__(obj=None, **kw_super)

    # # # DISPLAY # # #
    def __repr__(self):
        contents = [f'[depth={self.depth}, height={self.height}, size={self.size}]']
        return f'({self.__class__.__name__}({", ".join(contents)}))'

    def _shorthand_repr(self):
        '''returns shorthand repr for this node: "((depth, height, size)) ()".'''
        return f'(({self.depth}, {self.height}, {self.size})) ()'


class MovieOrganizerNode(EmptyMovieNode):
    '''makes an empty MoviePlotNode; can be used to improve tree readability.
    E.g. if there are 3 nodes for the same axis & of roughly equal importance,
        might want to have them all be children of a MovieOrganizerNode,
        instead of arbitrarily picking one of them to be the parent.

    name: str
        name for this node; to be displayed in __repr__.
    '''
    def __init__(self, name='', **kw_super):
        super().__init__(**kw_super)
        self.name = name

    # # # DISPLAY # # #
    def __repr__(self):
        contents = [f'[depth={self.depth}, height={self.height}, size={self.size}]']
        if self.name: contents.append(f'{self.name!r}')
        return f'--- {self.__class__.__name__}({", ".join(contents)}) ---'

    def _shorthand_repr(self):
        '''returns shorthand repr for this node (without children).
        returns '--- name ---' if self.name provided, else "((depth, height, size))".
        '''
        if self.name:
            return f'--- {self.name} ---'
        else:
            return f'(({self.depth}, {self.height}, {self.size})) {repr_simple(self.obj)}'



"""
File Purpose: the most basic of tests - import & initialize.

Also, clears global artifacts folder if it exists.
"""
import os

# BASIC TESTS
def test_import():
    '''ensure can import PlasmaCalcs'''
    import PlasmaCalcs as pc

def test_init():
    '''ensure can initialize some objects from PlasmaCalcs'''
    import PlasmaCalcs as pc
    p = pc.PlasmaCalculator()
    


"""
Package Purpose: helper classes for dimensions subpackage.

Not intended for direct use; use a subclass instead.


This is where most of the logic for dimensions is implemented. Overview:

    DimensionValue: a single value of a dimension.
        E.g. Fluid('H', 0) or Snap('1600', 20, t=1.6e-5, dt=1e-8).
        can be converted to string or int, and is hashable (can be used as dict keys).

    DimensionValueList: a list of DimensionValue elements.
        E.g. FluidList([Fluid('H', 0), Fluid('H+', 1), Fluid('C+', 2)])
        list-like, also provides DimensionValueList.get(key) to look up elements,
            by key = str, int, slice, range, tuple, list, or DimensionValue.

    DimPoint: a dict of values representing a single point in dimension-space.
        some parts of the code will refer to "dimpoints".

    Dimension: a single dimension, representing a current value AND a list of all possible values.
        The "current value" might be a list of multiple values! And, it can change frequently.
            Meanwhile, the list of all possible values will probably not change.
        You will probably create a subclass of Dimension before making a subclass DimensionHaver,
            and use Dimension.setup_haver to decorate the DimensionHaver subclass.
        See help(Dimension) for more details.
        And/or, see examples in snaps.py, fluids.py, components.py, runs.py.

    DimensionHaver: a class which can have multiple Dimensions attached to it.
        Manages multiple Dimensions and provides methods for working with them, such as:
            current_n_dimpoints, dim_values, enumerate_dimpoints, get_first_dimpoint.
        Additionally, provides the load_across_dims method.
            load_across_dims is extremely useful for proper bookkeeping when loading arrays from multiple files,
                so it will probably be most commonly used in the hookup's "BasesLoader".
                E.g., loading 'n' across 'snap' and 'fluid' because number density varies from snap to snap
                    and from fluid to fluid. load_across_dims handles iterating across dimension values
                    and concatenating the results. (The "load from file & label array" function must be
                    provided to load_across_dims; that is defined elsewhere. See e.g. load_maindims_var.)
            Examples: see EppicBasesLoader, EbysusBasesLoader.
            Note: it might not appear directly; the method was so commonly used in "BasesLoader" classes,
                that the relevant code for calling it was included as an option in quantity_tools.py.
                Providing @known_var(..., load_across_dims=[list of strings]) will lead to calling
                load_across_dims while using the decorated function. (similar for @known_pattern.)
"""
from .dim_point import DimPoint, DimRegion
from .dimension import Dimension
from .dimension_haver import DimensionHaver
from .dimension_value import (
    DimensionValue, DimensionValueList,
    SliceableList, string_int_lookup,
    DimensionSpecialValueSpecifier, DimensionSingleValueSpecifier,
    DimensionFractionalIndexer,
    SpecialDimensionValue, UniqueDimensionValue,
)
from .multi_slices import MultiSlices, _paramdocs_multi_slices



"""
File purpose: blur calculations (e.g. convolution with a gaussian kernel)

Futurewarning: "blur" might eventually be more generic than gaussian_filter,
    e.g. might add a blur_mode with 'gaussian' as one option,
    if there are any other good blurring methods to consider.
"""

from ..quantity_loader import QuantityLoader
from ...tools import (
    alias, simple_property,
    format_docstring,
    UNSET,
    xarray_gaussian_filter, xarray_ensure_dims,
)
from ...defaults import DEFAULTS


class BlurLoader(QuantityLoader):
    '''blur calculations (e.g. convolution with a gaussian kernel)'''
    @format_docstring(xarray_gaussian_filter_docs=xarray_gaussian_filter.__doc__, sub_ntab=2)
    def gaussian_filter(self, array, dim=UNSET, sigma=UNSET, **kw_xarray_gaussian_filter):
        '''xarray_gaussian_filter with defaults from self.blur_dims, self.blur_sigma.
        kwargs are passed to xarray_gaussian_filter.

        For convenience, docs for xarray_gaussian_filter are copied below:
        xarray_gaussian_filter_docs
        ---------------------------
            {xarray_gaussian_filter_docs}
        '''
        if dim is UNSET:
            dim = self.blur_dims
            # sets dim = existing dims only, and promotes any coords to dims if needed:
            array, dim = xarray_ensure_dims(array, dim,
                            missing_dims='ignore', return_existing_dims=True)
        if sigma is UNSET:
            sigma = self.blur_sigma
        return xarray_gaussian_filter(array, dim=dim, sigma=sigma, **kw_xarray_gaussian_filter)

    blur = alias('gaussian_filter')

    cls_behavior_attrs.register('blur_dims', getdefault=lambda ql: getattr(ql, 'maindims', []))
    @property
    def blur_dims(self):
        '''the dims over which to possibly apply blur (BlurLoader methods).
        will only blur along these dims for an array if they actually appear in the array.
        None --> use self.maindims. (this is the default.)
        '''
        if getattr(self, '_blur_dims', None) is None:
            return getattr(self, 'maindims', [])
        else:
            return self._blur_dims
    @blur_dims.setter
    def blur_dims(self, value):
        self._blur_dims = value

    cls_behavior_attrs.register('blur_sigma', default=DEFAULTS.GAUSSIAN_FILTER_SIGMA)
    blur_sigma = simple_property('_blur_sigma', setdefault=lambda: DEFAULTS.GAUSSIAN_FILTER_SIGMA,
        doc=f'''the default sigma to use for blurring calculations.
        None --> use DEFAULTS.GAUSSIAN_FILTER_SIGMA (default: {DEFAULTS.GAUSSIAN_FILTER_SIGMA})''')

    # # # LOADABLE QUANTITIES # # #
    @known_pattern(r'(blur|gaussian_filter)_(.+)', deps=[1])  # 'blur_{var}' or 'gaussian_filter_{var}'
    def get_blur(self, var, *, _match=None):
        '''gaussian_filter(var). Applied along all self.blur_dims in array.
        'blur_var' or 'gaussian_filter_var'; both are equivalent.
        '''
        _blurstr, var = _match.groups()
        array = self(var)
        return self.blur(array)

    @known_pattern(r'(blurt|gaussian_filtert)_(.+)', deps=[1])  # 'blurt_{var}' or 'gaussian_filtert_{var}'
    def get_blurt(self, var, *, _match=None):
        '''gaussian_filter(var), temporarily using self.blur_dims = ['snap']
        'blurt_var' or 'gaussian_filtert_var'; both are equivalent.
        '''
        _blurstr, var = _match.groups()
        return self(f'blur_{var}', blur_dims=['snap'])

    @known_pattern(r'(blurk|gaussian_filterk)_(.+)', deps=[1])  # 'blurk_{var}' or 'gaussian_filterk_{var}'
    def get_blurk(self, var, *, _match=None):
        '''gaussian_filter(var), temporarily using self.blur_dims = ['k_x', 'k_y', 'k_z']
        'blurk_var' or 'gaussian_filterk_var'; both are equivalent.
        '''
        _blurstr, var = _match.groups()
        return self(f'blur_{var}', blur_dims=['k_x', 'k_y', 'k_z'])



"""
File Purpose: EppicCalculator
"""

import numpy as np
import xarray as xr

from .eppic_addon_quantity_loader import EppicAddonQuantityLoader
from .eppic_bases import EppicBasesLoader
from .eppic_choose_params import EppicChooseParams, EppicTimescales
from .eppic_input_deck import EppicInputDeck
from .eppic_direct_loader import EppicDirectLoader
from .eppic_io_tools import read_eppic_snaps_info
from .eppic_moments import EppicMomentsLoader
from .eppic_plotters import EppicPlotterManager
from .eppic_runtime_info import EppicRuntimeInfoLoader
from .eppic_sim_info import EppicSimInfoLoader
from .eppic_subsampler import EppicSubsamplable
from ...dimensions import SnapList, INPUT_SNAP
from ...errors import DimensionError, FluidValueError, InputConflictError, InputError
from ...plasma_calculator import MultifluidPlasmaCalculator
from ...tools import (
    alias, alias_child,
    UNSET,
)
from ...defaults import DEFAULTS

class EppicCalculator(EppicBasesLoader,
                      EppicMomentsLoader, EppicRuntimeInfoLoader, EppicSimInfoLoader,
                      EppicSubsamplable, EppicDirectLoader,
                      EppicPlotterManager,
                      EppicAddonQuantityLoader,
                      MultifluidPlasmaCalculator):
    '''PlasmaCalculator for Eppic outputs.
    To initialize from an eppic.i file, use EppicCalculator.from_here().
    
    Use EppicCalculator.help() for more help.

    input_deck: EppicInputDeck
        the EppicInputDeck which is being used by this calculator.
    snaps_from: 'parallel' or 'timers'
        how to determine existing snaps.
        'parallel' --> from files in directory 'parallel' (located adjacent to 'eppic.i')
        'timers' --> from data in 'domain000/timers.dat' ('domain000' adjacent to 'eppic.i')
    u: None or UnitsManager
        if provided, set self.u = u, and ignore u_l, u_t, u_n, and ne_si.
    u_l, u_t, u_n, ne_si: None or value
        must provide one of these values in order to fully determine the units.
        u_l: length [si] = u_l * length [raw]
        u_t: time [si] = u_t * time [raw]
        u_n: number density [si] = u_n * number density [raw]
        ne_si: electron number density [si]. (ne [raw] will be loaded from input_deck)
    kw_units: dict
        any additional values used to determine the units.
        can also provide u_l, u_t, u_n, or ne_si, here, instead of as direct kwarg.
        See help(UnitsManagerPIC.from_pic) for more details.
    quasineutral: None or bool
        if provided, set self.quasineutral.
        otherwise, infer it from self.input_deck.is_quasineutral().
        in quasineutral mode, distribution 0 is assumed to be electrons,
            and different routines are used to calculate its density & other parameters.
    _more_init_checks: bool, default False
        if True, perform additional checks during initialization.
        False by default for slight efficiency improvement,
            and to avoid possibility of init checks failing but object still being usable.

    other kwargs are passed to super().__init__.
    '''
    # # # DEFAULTS # # #
    _quasineutral = False

    # # # CREATION / INITIALIZATION # # #
    def __init__(self, input_deck, *, snaps_from='parallel',
                 u=None, u_l=None, u_t=None, u_n=None, ne_si=None, kw_units=dict(),
                 quasineutral=None,
                 _more_init_checks=False,
                 **kw_super):
        self.input_deck = input_deck
        kwu = kw_units.copy()
        self.init_units_manager(u=u, u_l=kwu.pop('u_l', u_l), u_t=kwu.pop('u_t', u_t),
                                u_n=kwu.pop('u_n', u_n), ne_si=kwu.pop('ne_si', ne_si), **kwu)
        self.init_maindims()
        self.init_fluids()
        kw_super.setdefault('jfluid', 0)  # if jfluid not provided, use jfluid=0.
        self.init_snaps(snaps_from=snaps_from)
        if quasineutral is None: quasineutral = self.input_deck.is_quasineutral()
        super().__init__(input_deck, quasineutral=quasineutral, **kw_super)
        if _more_init_checks:
            self._check_files_readable()

    @classmethod
    def from_here(cls, filename='eppic.i', *, dist_names=None, **kw):
        '''create EppicCalculator from input deck file found here (at filename).
        dist_names: None or dict
            {N: name for distribution N} for any number of distributions in filename.
            E.g.: {0: 'e', 1: 'H+', 2: 'C+'}
        '''
        __tracebackhide__ = DEFAULTS.TRACEBACKHIDE
        return cls(EppicInputDeck.from_file(filename, dist_names=dist_names), **kw)

    # # # INFO # # #
    dirname = alias_child('input_deck', 'dirname')
    filename = alias_child('input_deck', 'filename')

    # # # UNITS # # #
    def init_units_manager(self, *, u=None, u_l=None, u_t=None, u_n=None, ne_si=None, **kw_get_units_manager):
        '''set units manager self.u, based on input_deck.
        must provide one of these values in order to fully determine the units.
            u: UnitsManager object. if provided, set self.u = u.
            u_l: length [si] = u_l * length [raw]
            u_t: time [si] = u_t * time [raw]
            u_n: number density [si] = u_n * number density [raw]
            ne_si: electron number density [si]. (ne [raw] will be loaded from input_deck)
        '''
        __tracebackhide__ = DEFAULTS.TRACEBACKHIDE
        if u is None:
            u = self.input_deck.get_units_manager(u_l=u_l, u_t=u_t, u_n=u_n, ne_si=ne_si, **kw_get_units_manager)
        self.u = u

    # # # DIMENSIONS SETUP -- MAINDIMS # # #
    def init_maindims(self):
        '''set self.maindims based on input_deck.'''
        __tracebackhide__ = DEFAULTS.TRACEBACKHIDE
        self.maindims = self.input_deck.maindims()

    def get_maindims_coords(self, units=None):
        '''return dict of {'x': xcoords, 'y': ycoords}. Possibly also 'z': zcoords.
        units: None or str
            result will be in this unit system. if None, use self.coords_units (default: self.units).
        coords will be subsampled appropriately if subsampling_info exists.
        coords will also be sliced according to self.slices.
        '''
        result = self.input_deck.get_space_coords()
        u_l = self.u('l', self.coords_units_explicit if units is None else units)
        result = {key: u_l * value for key, value in result.items()}
        result = self._apply_subsampling_to_maindims_coords(result)
        return self._apply_maindims_slices_to_dict(result)

    get_space_coords = alias('get_maindims_coords')

    # # # DIMENSIONS SETUP -- OTHERS # # #
    def init_snaps(self, *, snaps_from='parallel'):
        '''set self.snaps based on input_deck AND reading files.
        snaps_from: 'parallel' or 'timers'
            how to determine existing snaps.
            'parallel' --> from files in directory 'parallel' (located adjacent to 'eppic.i')
            'timers' --> from data in 'domain000/timers.dat' ('domain000' adjacent to 'eppic.i')
        '''
        __tracebackhide__ = DEFAULTS.TRACEBACKHIDE
        self._snaps_from = snaps_from
        dt = self.input_deck["dt"]
        dir_ = self.input_deck.dirname
        names, times = read_eppic_snaps_info(dir_, read_mode=self.input_deck["hdf_output_arrays"], dt=dt, snaps_from=snaps_from)
        self.snaps = SnapList.from_lists(s=names, t=times)

    @property
    def snaps_from(self):
        '''how existing self.snaps were determined. Probably 'parallel' or 'timers'.
        'parallel' --> from files in directory 'parallel' (located adjacent to 'eppic.i')
        'timers' --> from data in 'domain000/timers.dat' ('domain000' adjacent to 'eppic.i')

        setting self.snaps_from=value is equivalent to calling self.init_snaps(snaps_from=value).
        '''
        return self._snaps_from
    @snaps_from.setter
    def snaps_from(self, value):
        self.init_snaps(snaps_from=value)

    def init_fluids(self):
        '''set self.fluids and self.jfluids based on input_deck.'''
        __tracebackhide__ = DEFAULTS.TRACEBACKHIDE
        self.fluids = self.input_deck.dists
        self.jfluids = self.input_deck.neutrals

    # [TODO][MV] should this method go somewhere else?
    def assign_velocity_coords(self, array, fluid=UNSET):
        '''assign velocity dims and coords, based on fluid.get_velocity_coords().
        E.g., for "vdist" data. Array shape and coords inferred from input deck.
        (If subsampling_info exists, that will also be applied appropriately.)
        returns an xarray with proper details for PlasmaCalcs. Output units according to self.units.

        This is not like the assign_{dim}_coord functions, which assign 0D coord to an existing xarray;
        this method creates a *new* xarray based on array, and velocity dims & coords are >0 dimensional.

        array: 3D array
            probably the "vdist" data, e.g. directly from reading output data from one snapshot.
        fluid: UNSET, int, str, Fluid, or other way to specify a single fluid.
            UNSET --> use self.fluid.
            else --> temporarily set self.fluid = fluid, during this operation.
        '''
        if np.ndim(array) != 3:
            raise DimensionError(f'array must be 3D, but got ndim={np.ndim(array)}')
        if fluid is UNSET:
            fluid = self.fluid
        with self.using(fluid=fluid):
            if self.current_n_fluid() != 1:
                raise FluidValueError(f'assign_velocity_coords expects 1 fluid, got {self.current_n_fluid()}')
            fluid = self.fluid_list()[0]  # convert to "definitely a single fluid, not a list of length 1"
            vcoords = fluid.get_velocity_coords(units=self.u('speed'))
            vcoords = self._apply_subsampling_to_vdist_coords(vcoords, fluid)
        result = xr.DataArray(array, coords=vcoords, dims=('vx', 'vy', 'vz'))
        return result

    # # # DISPLAY # # #
    def _repr_contents(self):
        '''return list of strings to use inside self.__repr__'''
        contents = super()._repr_contents()
        contents = [f'input_deck={self.input_deck!r}'] + list(contents)
        return contents

    # # # MANIPULATING / DIRECTLY USING INPUT DECK # # #
    def set_vars_from_inputs(self, snap=INPUT_SNAP, *, _version='new', **kw):
        '''set self.snap to INPUT_SNAP. self will automatically load eppic.i values when snap=INPUT_SNAP.

        (This function was useful before load_input() was implemented (on 2025-04-25),
            but now that load_input() exists, there's probably not a compelling reason to keep this around...
            However, we will keep it around for a while, in case someone was using it.
            It *should* also be *mostly* backwards compatible, as setting self.snap = INPUT_SNAP
                will now produce equivalent result as the old version would have produced,
                except the new version doesn't touch self.setvars in order to do it.

        snap: INPUT_SNAP or any snap specifier
            sets self.snap = snap, and sets values at this snap. (doesn't restore old self.snap after.)
            INPUT_SNAP --> special value; refers to "snap from input deck, not a real snap".
            You can later access the set values by using self.snap = PlasmaCalcs.INPUT_SNAP
            if _version=='new', snap MUST be INPUT_SNAP, else will crash with InputConflictError
        _version: 'new' or 'old'
            if 'new', just set self.snap = INPUT_SNAP.
            if 'old', set self.snap=snap, then use self.set_bases() to adjust self.setvars appropriately.

        kw are passed as self.using(**kw) while using this method.
        '''
        if _version == 'new':
            if snap != INPUT_SNAP:
                raise InputConflictError(f'if _version="new", snap must be INPUT_SNAP, but got {snap!r}')
            self.snap = INPUT_SNAP
        elif _version == 'old':
            self.snap = snap
            with self.using(**kw, units='raw'):  # input deck uses 'raw' units. setvars handles units later.
                # fluid-specific
                for fluid in self.iter_fluids():
                    n = fluid.get_n0()
                    vx, vy, vz = fluid.get_v0()
                    try:
                        vtherm = fluid.get_vth0(x=...)
                    except AssertionError:
                        raise NotImplementedError('set_vars_from_inputs for non-isotropic vtherm')
                    self.set_bases(n=n, vx=vx, vy=vy, vz=vz)
                    self.set('vtherm', vtherm)
                # global
                self.set('phi', 0)  # use phi=0 for initial timestep (getting E will give E_ext.)
        else:
            raise InputError(f'unknown _version={_version!r}. Must be "new" or "old".')

    def choose_params(self, **kw_init):
        '''returns EppicChooseParams based on self.

        Common usage:
            # write a new eppic.i file based on the eppic.i file in self.dirname,
            # and using the current values of params from self.
            self.choose_params().write(dst='eppic_updated.i')

        uses self.dirname and self.get_vals_for_inputs().

        kwargs go to EppicChooseParams.__init__
        '''
        return EppicChooseParams(self, **kw_init)

    def timescales(self, **kw_init):
        '''returns EppicTimescales based on self.
        CAUTION: result will have units of self.units;
            remember to switch to 'raw' when comparing to input_deck.
        '''
        return EppicTimescales(self, **kw_init)

